import click
from crypto_strategy.strategies.ma_strategy import BestMaStrategy, CheckMaIndicators
from crypto_strategy.strategies.bo_strategy import BestBoStrategy, CheckBoIndicators
from crypto_strategy.reporting.report import generate_report as _generate_report


SYMBOLS = [
    'BTC', 'ETH', 'BNB', 'SOL', 'ADA',
    'XRP', 'DOT', 'DOGE', 'AVAX', 'LUNA',
    'LTC', 'UNI', 'LINK', 'ALGO',
    'VET', 'MATIC',
    # # under review
    # 'XTZ', 'BCH', 'EGLD', 'XLM', 'AXS',
]
BASES = ['USDT']
SYMBOLS = [s + b for s in SYMBOLS for b in BASES]


@click.group()
def cli():
    pass


@cli.command()
@click.option('--symbol', '-s', type=str, default=None, help="asset name, e.g., BTCUSDT")
@click.option('--freq', '-f', required=True, type=click.Choice(['4h', '1h']), help="frquency to use")
@click.option('--res_dir', '-r', required=True, type=str, help="directory for outputs")
@click.option('--flag_filter', '-g', type=str, default=None, show_default=True, help='flag to use, mmi | ang')
@click.option('--flag_stop', '-t', type=str, default=None, show_default=True, help='early stop flag, ts_stop | sl_stop | tp_stop')
def best_ma_strategy(symbol, freq, res_dir, flag_filter, flag_stop):
    parts = res_dir.split('-')
    if freq not in parts:
        raise ValueError('Mismatch found in freq setting and output dir')
    if (flag_filter and flag_filter not in parts) or (not flag_filter and 'filter' in parts):
        raise ValueError('Mismatch found in filter setting and output dir')
    if flag_stop and 'stop' not in res_dir:
        raise ValueError('Mismatch found in flag_stop setting and output dir')
    symbols = symbol if symbol else SYMBOLS
    BestMaStrategy(symbols, freq, res_dir, flag_filter, flag_stop)


@cli.command()
@click.option('--symbol', '-s', type=str, default=None, help='asset name, e.g., BTCUSDT')
@click.option('--date', '-d', type=str, help='the date when the results are generated')
@click.option('--res_dir', '-r', required=True, type=str, help='directory for outputs')
@click.option('--flag_filter', '-g', type=str, default=None, show_default=True, help='flag to use, mmi | ang')
@click.option('--flag_stop', '-t', type=str, default=None, show_default=True, help='early stop flag, ts_stop | sl_stop | tp_stop')
def check_ma_indicators(symbol, date, res_dir, flag_filter, flag_stop):
    parts = res_dir.split('-')
    if (flag_filter and flag_filter not in parts) or (not flag_filter and 'filter' in parts):
        raise ValueError('Mismatch found in filter setting and output dir')
    if 'ma' not in parts:
        raise ValueError('Make sure the res_dir is for MA strategy')
    if flag_stop and 'stop' not in res_dir:
        raise ValueError('Mismatch found in flag_stop setting and output dir')
    symbols = symbol if symbol else SYMBOLS
    CheckMaIndicators(symbols, date, res_dir, flag_filter, flag_stop)


@cli.command()
@click.option('--symbol', '-s', type=str, default=None, show_default=True, help="asset name, e.g., BTCUSDT")
@click.option('--freq', '-f', required=True, type=click.Choice(['4h', '1h']), help="frquency to use")
@click.option('--res_dir', '-r', required=True, type=str, help="directory for outputs")
@click.option('--flag_filter', '-g', type=str, default=None, show_default=True, help='filter to use, vol | ang')
@click.option('--flag_stop', '-t', type=str, default=None, show_default=True, help='early stop flag, ts_stop | sl_stop | tp_stop')
def best_bo_strategy(symbol, freq, res_dir, flag_filter, flag_stop):
    parts = res_dir.split('-')
    if freq not in parts:
        raise ValueError('Mismatch found in freq setting and output dir')
    if (flag_filter and flag_filter not in parts) or (not flag_filter and 'filter' in parts):
        raise ValueError('Mismatch found in filter setting and output dir')
    if 'bo' not in parts:
        raise ValueError('Mismatch found in strategy bo and output dir')
    if flag_stop and 'ts_stop' not in parts:
        raise ValueError('Mismatch found in ts_stop setting and output dir')
    symbols = symbol if symbol else SYMBOLS
    BestBoStrategy(symbols, freq, res_dir, flag_filter, flag_stop)


@cli.command()
@click.option('--symbol', '-s', type=str, default=None, help="asset name, e.g., BTCUSDT")
@click.option('--date', '-d', type=str, help='the date when the results are generated')
@click.option('--res_dir', '-r', required=True, type=str, help="directory for outputs")
@click.option('--flag_filter', '-g', type=str, default=None, show_default=True, help='filter to use, vol | ang')
@click.option('--flag_stop', '-t', type=str, default=None, show_default=True, help='early stop flag, ts_stop | sl_stop | tp_stop')
def check_bo_indicators(symbol, date, res_dir, flag_filter, flag_stop):
    parts = res_dir.split('-')
    if (flag_filter and flag_filter not in parts) or (not flag_filter and 'filter' in parts):
        raise ValueError('Mismatch found in filter setting and output dir')
    if 'bo' not in parts:
        raise ValueError('Mismatch found in strategy bo and output dir')
    if flag_stop and 'ts_stop' not in parts:
        raise ValueError('Mismatch found in ts_stop setting and output dir')
    symbols = symbol if symbol else SYMBOLS
    CheckBoIndicators(symbols, date, res_dir, flag_filter, flag_stop)


@cli.command()
@click.option('--symbol', '-s', type=str, default=None, help="asset name, e.g., BTCUSDT")
def generate_report(symbol):
    if symbol:
        _generate_report(symbol)
    else:
        for symb in SYMBOLS:
            _generate_report(symb)


if __name__ == '__main__':
    cli()



import math
import itertools as itr
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from statsmodels.distributions.empirical_distribution import ECDF

sharpe_ratio = lambda r: r.mean() / (r.std()+0.0000001) * (365 ** 0.5)


class CSCV(object):
    """Combinatorially symmetric cross-validation algorithm.
    Calculate backtesting about overfitting probability distribution and performance degradation.

    Attributes:
        n_bins:A int of CSCV algorithm bin size to control overfitting calculation.Default is 10.
        objective:A function of in sample(is) and out of sample(oos) return benchmark algorithm.Default is lambda r:r.mean().

    """
    def __init__(self, n_bins=10, objective=sharpe_ratio):
        self.n_bins = n_bins
        self.objective = objective
        self.bins_enumeration = [set(x) for x in itr.combinations(np.arange(10), 10 // 2)]

        self.Rs = [pd.Series(dtype=float) for i in range(len(self.bins_enumeration))]
        self.R_bars = [pd.Series(dtype=float) for i in range(len(self.bins_enumeration))]

    def add_daily_returns(self, daily_returns):
        """Add daily_returns in algorithm.

        Args:
          daily_returns: A dataframe of trading daily_returns.

        """
        bin_size = daily_returns.shape[0] // self.n_bins
        bins = [daily_returns.iloc[i*bin_size: (i+1) * bin_size] for i in range(self.n_bins)]

        for set_id, is_set in enumerate(self.bins_enumeration):
            oos_set = set(range(10)) - is_set
            is_returns = pd.concat([bins[i] for i in is_set])
            oos_returns = pd.concat([bins[i] for i in oos_set])
            R = self.objective(is_returns)
            R_bar = self.objective(oos_returns)
            self.Rs[set_id] = self.Rs[set_id].append(R)
            self.R_bars[set_id] = self.R_bars[set_id].append(R_bar)

    def estimate_overfitting(self, plot=False):
        """Estimate overfitting probability.

        Generate the result on Combinatorially symmetric cross-validation algorithm.
        Display related analysis charts.

        Args:
          plot: A bool of control plot display. Default is False.

        Returns:
          A dict of result include:
          pbo_test: A float of overfitting probability.
          logits: A float of estimated logits of OOS rankings.
          R_n_star: A list of IS performance of th trategies that has the best ranking in IS.
          R_bar_n_star: A list of find the OOS performance of the strategies that has the best ranking in IS.
          dom_df: A dataframe of optimized_IS, non_optimized_OOS data.

        """
        # calculate strategy performance in IS(R_df) and OOS(R_bar_df)
        R_df = pd.DataFrame(self.Rs)
        R_bar_df = pd.DataFrame(self.R_bars)

        # calculate ranking of the strategies
        R_rank_df = R_df.rank(axis=1, ascending=False, method='first')
        R_bar_rank_df = R_bar_df.rank(axis=1, ascending=False, method='first')

        # find the IS performance of th trategies that has the best ranking in IS
        r_star_series = (R_df * (R_rank_df == 1)).unstack().dropna()
        r_star_series = r_star_series[r_star_series != 0].sort_index(level=-1)

        # find the OOS performance of the strategies that has the best ranking in IS
        r_bar_star_series = (R_bar_df * (R_rank_df == 1)).unstack().dropna()
        r_bar_star_series = r_bar_star_series[r_bar_star_series != 0].sort_index(level=-1)

        # find the ranking of strategies which has the best ranking in IS
        r_bar_rank_series = (R_bar_rank_df * (R_rank_df == 1)).unstack().dropna()
        r_bar_rank_series = r_bar_rank_series[r_bar_rank_series != 0].sort_index(level=-1)

        # probability of overfitting

        # estimate logits of OOS rankings
        logits = (1-((r_bar_rank_series)/(len(R_df.columns)+1))).map(lambda p: math.log(p/(1-p)))
        # prob = (logits < 0).sum() / len(logits)

        # stochastic dominance

        # caluclate
        if len(r_bar_star_series) != 0:
            y = np.linspace(
                min(r_bar_star_series), max(r_bar_star_series), endpoint=True, num=1000
            )

            # build CDF performance of best candidate in IS
            R_bar_n_star_cdf = ECDF(r_bar_star_series.values)
            optimized = R_bar_n_star_cdf(y)

            # build CDF performance of average candidate in IS
            R_bar_mean_cdf = ECDF(R_bar_df.median(axis=1).values)
            non_optimized = R_bar_mean_cdf(y)

            dom_df = pd.DataFrame(
                dict(optimized_IS=optimized, non_optimized_OOS=non_optimized), index=y)
            dom_df["SD2"] = -(dom_df.non_optimized_OOS - dom_df.optimized_IS).cumsum()
        else:
            dom_df = pd.DataFrame(columns=['optimized_IS', 'non_optimized_OOS', 'SD2'])

        ret = {
            'pbo_test': (logits < 0).sum() / len(logits),
            'logits': logits.to_list(),
            'R_n_star': r_star_series.to_list(),
            'R_bar_n_star': r_bar_star_series.to_list(),
            'dom_df': dom_df,
        }

        if plot:
            # probability distribution
            plt.title('Probability Distribution')
            plt.hist(x=[r for r in ret['logits'] if r > -10000], bins='auto')
            plt.xlabel('Logits')
            plt.ylabel('Frequency')
            plt.show()

            # performance degradation
            plt.title('Performance degradation')
            plt.scatter(ret['R_n_star'], ret['R_bar_n_star'])
            plt.xlabel('In-sample Performance')
            plt.ylabel('Out-of-sample Performance')

            # first and second Stochastic dominance
            plt.title('Stochastic dominance')
            ret['dom_df'].plot(secondary_y=['SD2'])
            plt.xlabel('Performance optimized vs non-optimized')
            plt.ylabel('Frequency')
            plt.show()

        return ret



import os
import time
import pandas as pd
from finlab_crypto.crawler import get_all_binance


def save_stats(stats: pd.DataFrame, output_path, filename) -> None:
    filename = join_path(output_path, filename)
    stats.to_pickle(filename)


def get_acc_returns(daily_returns):
    acc_returns = {
        'Ret [:21-04-14]': (daily_returns[:'2021-04-15'] + 1).cumprod()[-1],
        'Ret [21-04-15:21-07-20]': (daily_returns['2021-04-15':'2021-07-21'] + 1).cumprod()[-1],
        'Ret [21-07-21:21-11-10]': (daily_returns['2021-07-21':'2021-11-11'] + 1).cumprod()[-1],
        'Ret [21-11-11:]': (daily_returns['2021-11-11':] + 1).cumprod()[-1]
    }
    return acc_returns


def check_and_create_dir(dname, *args) -> str:
    dname = os.path.join(dname, *args)
    has_dir = os.path.isdir(dname)
    if not has_dir:
        os.mkdir(dname)
    return dname


def join_path(output_path, filename) -> str:
    return os.path.join(output_path, filename)


def download_crypto_history(symbol, freq) -> pd.DataFrame:
    history = pd.DataFrame()
    try:
        print('try', symbol)
        history = get_all_binance(symbol, freq)
    except Exception as e:
        print('fail', e)
    return history


def download_all_portfolio_data(symbols: list, freq: list, dir_history='history') -> dict:
    check_and_create_dir(dir_history)
    ohlcvs = {}
    for s, f in zip(symbols, freq):
        time.sleep(1)
        ohlcvs[(s, f)] = download_crypto_history(s, f)
    return ohlcvs



import time
import datetime
import warnings
import pandas as pd
import plotly.express as px
from datetime import timezone
from binance.enums import *
from binance.client import Client
from IPython.display import display
from finlab_crypto.crawler import get_nbars_binance, get_all_binance


warnings.filterwarnings(
    "ignore",
    message="The localize method is no longer necessary.*",
)


class TickerInfo():
    """Ticker basic info.

    Get asset amount and convert price to BTC .

    Attributes:
        client: A Binance client object where api_key, api_secret is required.

    """
    def __init__(self, client):
        self.exinfo = client.get_exchange_info()
        self.info = client.get_account()
        self.tickers = client.get_symbol_ticker()

    @staticmethod
    def _list_select(list, key, value):
        ret = [l for l in list if l[key] == value]
        if len(ret) == 0:
            return None
        else:
            return ret[0]

    def get_base_asset(self, symbol):
        """Get base asset data of a given symbol.

        Args:
          symbol: A str of trading target name.

        Returns:
            A str of base asset (ex: 'BTC').
        """
        sinfo = self._list_select(self.exinfo['symbols'], 'symbol', symbol)
        return sinfo['baseAsset']

    def get_quote_asset(self, symbol):
        """Get quote asset data of a given symbol.

        Args:
          symbol: A str of trading target name.

        Returns:
          A float of quote asset.
        """
        sinfo = self._list_select(self.exinfo['symbols'], 'symbol', symbol)
        return sinfo['quoteAsset']

    def get_asset_price_in_btc(self, asset):
        """Convert price to BTC .

        Args:
          asset: A str of asset name (ex: 'ETH').

        Returns:
          A float of price in BTC.
        """
        if asset == 'BTC':
            return 1

        ret = self._list_select(self.tickers, 'symbol', asset + 'BTC')

        if ret is not None:
            return float(ret['price'])

        ret = self._list_select(self.tickers, 'symbol', 'BTC' + asset)
        if ret is not None:
            return 1 / float(ret['price'])

        return None


class TradingMethod():
    """Trading method in online init setting.

    Create trading method object for TradingPortfolio register .

    Attributes:
        symbols: A list of trading pair (ex: ['USDTBTC','ETHBTC']).
        freq: A str of trading time period (ex: '4h').
        lookback: An int of the length of historical data (ex:1000).
        strategy: A function that is your customized strategy (ex:trend_strategy).
        variables: A dict of your customized strategy attributes (ex:dict(name='sma',n1=30,n2=130,),).
        weight_btc: A float of btc for each commodity operation (ex: 0.2).
        filters: A dict that is your customized filter (ex:{}).
        name: A str of your trading method name (ex:'altcoin-trend-hullma').

    """
    def __init__(self, symbols, freq,
                 lookback, strategy, variables,
                 weight_btc=None, weight=None, weight_unit=None, filters=None, name='', execution_price='close'):
        self.symbols = symbols
        self.freq = freq
        self.lookback = lookback
        self.strategy = strategy
        self.variables = variables
        self.weight_btc = weight_btc
        self.weight = weight
        self.weight_unit = weight_unit
        self.filters = filters
        self.name = name
        self.execution_price = execution_price

        if self.weight_btc is None and self.weight is None:
            raise Exception("weight_btc or weight is missing.")

        if self.weight_btc is not None and self.weight is not None:
            raise Exception("weight_btc and weight should not be assigned at the same time")

        if self.weight_btc:
            self.weight = self.weight_btc
            self.weight_unit = 'BTC'


class TradingPortfolio():
    """Connect Binance account.

    The core class to connect Binance  with API, in order to connect account info,
    register strategt.

    Attributes:
        binance_key: A str of is binance authorization key.
        binance_secret: A str of is binance authorization secret.

    """
    def __init__(self, binance_key, binance_secret, execute_before_candle_complete=False):
        self._client = Client(api_key=binance_key, api_secret=binance_secret)
        self._trading_methods = []
        self._margins = {}
        self.ticker_info = TickerInfo(self._client)
        self.default_stable_coin = 'USDT'
        self.execute_before_candle_complete = execute_before_candle_complete

    def set_default_stable_coin(self, token):
        self.default_stable_coin = token

    def register(self, trading_method):
        """Rigister TradingMethod object.
        Args:
          trading_method: A object of TradingMethod().
        """

        if trading_method.execution_price == 'open' and self.execute_before_candle_complete:
            raise Exception("Detect execute_before_candle_complete=True and trading_method.execution_price is open",
                            "Please set trading_method.execute_before_candle_complete to False",
                            " and execute live trading right after candles are complete.")

        self._trading_methods.append(trading_method)

    def register_margin(self, asset, weight_btc):
        """Rigister weight_btc as operation amount.
        Args:
          asset: A str of asset name (ex: 'USDT')
          weight_btc: A float of btc for each commodity operation (ex: 0.2)
        """
        self._margins[asset] = weight_btc

    def get_all_symbol_lookback(self):
        """Get all symbol lookback.
        Use in get_ohlcvs(self) function.
        Returns:
            A dict of OHLCV lookback.

        """
        symbol_lookbacks = {}
        addition = {}

        weight_units = set()
        max_lookback = 0

        for method in self._trading_methods:
            weight_units.add((method.weight_unit, method.freq))
            max_lookback = max(max_lookback, method.lookback)
            for a in method.symbols:

                quote_asset = self.ticker_info.get_quote_asset(a)
                base_asset = self.ticker_info.get_base_asset(a)
                if (a, method.freq) not in symbol_lookbacks or method.lookback > symbol_lookbacks[(a, method.freq)]:
                    symbol_lookbacks[(a, method.freq)] = method.lookback

                if base_asset != method.weight_unit:
                    new_symbol = base_asset + method.weight_unit
                    if (new_symbol, method.freq) not in addition or method.lookback > addition[(new_symbol, method.freq)]:
                        addition[(new_symbol, method.freq)] = method.lookback

        for w, f in weight_units:
            if w != 'BTC':
                addition[('BTC' + w, f)] = max_lookback

        # add quote asset historical data
        # for (symbol, freq), lookback in symbol_lookbacks.items():
        #     base_asset = self.ticker_info.get_base_asset(symbol)
        #     if base_asset != self.quote_asset:
        #         new_symbol = base_asset + self.quote_asset
        #         addition[(new_symbol, freq)] = lookback

        return {**symbol_lookbacks, **addition}

    def get_ohlcvs(self):
        """Getting histrical price data through binance api.

        Returns:
            A DataFrame of OHLCV data , the number of data length is lookback.

        """
        symbol_lookbacks = self.get_all_symbol_lookback()

        ohlcvs = {}
        for (symbol, freq), lookback in symbol_lookbacks.items():
            ohlcvs[(symbol, freq)] = get_nbars_binance(symbol, freq, lookback, self._client)

        return ohlcvs

    def get_full_ohlcvs(self):
        """Getting all histrical price data through binance api.

        Returns:
            A DataFrame of OHLCV data for all.

        """
        symbol_lookbacks = self.get_all_symbol_lookback()
        ohlcvs = {}

        for (symbol, freq), lookback in symbol_lookbacks.items():
            ohlcvs[(symbol, freq)] = get_all_binance(symbol, freq)
            time.sleep(3)
        return ohlcvs

    def get_latest_signals(self, ohlcvs, html=False):
        """Get latest signals dataframe.

        Choose which strategy to implement on widgets GUI.

        Args:
          ohlcvs: A dataframe of symbel.
          html: A bool of controlling html generation.

        Returns:
          A dataframe of latest_signals data,
          The last_signals column is bool value of whether to execute the transaction.
          The value_in_btc column is present value of assets.
        """
        ret = []
        for method in self._trading_methods:
            for symbol in method.symbols:
                ohlcv = ohlcvs[(symbol, method.freq)].copy()

                # remove incomplete candle
                if self.execute_before_candle_complete is False and method.execution_price == 'close':
                    t = datetime.datetime.utcnow().replace(tzinfo=timezone.utc)
                    delta_t = ohlcv.index[-1] - ohlcv.index[-2]
                    ohlcv = ohlcv.loc[:t-delta_t]

                htmlname = f'{symbol}-{method.freq}-{method.name}.html' if html else None
                result = method.strategy.backtest(ohlcv,
                                                  method.variables, filters=method.filters, plot=html,
                                                  html=htmlname,
                                                  freq=method.freq, fees=0., slippage=0., execution_price=method.execution_price)

                signal = result.cash().iloc[-1] == 0
                return_ = 0

                # find weight if it is in the nested dictionary
                weight = method.weight
                if isinstance(weight, dict):
                    weight = (weight[symbol] if symbol in weight else weight['default'])

                entry_price = 0
                entry_time = 0
                value_in_btc = 0
                trade_price_type = method.execution_price
                if signal:
                    txn = result.positions.records
                    rds = result.orders.records
                    return_ = ohlcv[trade_price_type].iloc[-1] / rds['price'].iloc[-1] - 1
                    entry_price = rds['price'].iloc[-1]
                    entry_time = ohlcv.index[int(rds.iloc[-1]['idx'])]

                    base_asset = self.ticker_info.get_base_asset(symbol)
                    quote_asset = self.ticker_info.get_quote_asset(symbol)

                    if base_asset != method.weight_unit:
                        quote_symbol = base_asset + method.weight_unit
                        quote_history = ohlcvs[(quote_symbol, method.freq)]
                        quote_asset_price_previous = quote_history[trade_price_type].loc[entry_time]
                        quote_asset_price_now = quote_history[trade_price_type].iloc[-1]
                    else:
                        quote_asset_price_previous = 1
                        quote_asset_price_now = 1

                    if method.weight_unit != 'BTC':
                        btc_quote_price_previous = ohlcvs[('BTC' + method.weight_unit, method.freq)][trade_price_type].loc[entry_time]
                        btc_quote_price_now = ohlcvs[('BTC' + method.weight_unit, method.freq)][trade_price_type].iloc[-1]
                    else:
                        btc_quote_price_previous = 1
                        btc_quote_price_now = 1

                    previous_weight_btc = (weight / btc_quote_price_previous)
                    value_in_btc = previous_weight_btc / quote_asset_price_previous * quote_asset_price_now /\
                                (btc_quote_price_now / btc_quote_price_previous)
                    weight_btc = previous_weight_btc
                    previous_price_btc = quote_asset_price_previous / btc_quote_price_previous
                    present_price_btc = quote_asset_price_now / btc_quote_price_now
                    amount = previous_weight_btc / previous_price_btc
                    present_amount = value_in_btc / present_price_btc
                    assert 0.9999 < amount / present_amount < 1.0001

                else:
                    if method.weight_unit != 'BTC':
                        btc_quote_price_now = ohlcvs[('BTC' + method.weight_unit, method.freq)][trade_price_type].iloc[-1]
                    else:
                        btc_quote_price_now = 1

                    weight_btc = weight / btc_quote_price_now
                    amount = 0

                ret.append({
                    'symbol': symbol,
                    'method name': method.name,
                    'latest_signal': signal,
                    'weight_btc': weight_btc,
                    'freq': method.freq,
                    'return': return_,
                    'amount': amount,
                    'value_in_btc': value_in_btc * signal,
                    'latest_price': ohlcv[trade_price_type].iloc[-1],
                    'entry_price': entry_price,
                    'entry_time': entry_time,
                    'html': htmlname,
                })

        ret = pd.DataFrame(ret)
        return ret

    def calculate_position_size(self, signals, rebalance_threshold=0.03, excluded_assets=list()):
        """Calculate the proportion of asset orders.

        Calculate data is based on latest signals dataframe.

        Args:
          signals: A dataframe of signals.
          rebalance_threshold: A float of rebalance_threshold.
          excluded_assets: A list of asset name which are excluded calculation.

        Returns:
          diff_value: A dataframe of how many assets to deposit for each cryptocurrency.
          diff_value_btc: A dataframe of converting cryptocurrency to BTC.
          transaction: A dataframe of transaction(new order) data.
        """
        if self.default_stable_coin not in excluded_assets:
            excluded_assets.append(self.default_stable_coin)

        signals['base_asset'] = signals.symbol.map(self.ticker_info.get_base_asset)
        signals['quote_asset'] = signals.symbol.map(self.ticker_info.get_quote_asset)
        signals['base_value_btc'] = signals.latest_signal * signals.value_in_btc
        signals['quote_value_btc'] = -(signals.latest_signal.astype(int) * signals.weight_btc)

        quote_asset_list = list(set(signals.quote_asset))

        # calculate base and quote assets (in btc term)
        base_asset_value = pd.Series(signals.base_value_btc.values, index=signals.base_asset)
        quote_asset_value = pd.Series(signals.quote_value_btc.values, index=signals.quote_asset)
        base_asset_value = base_asset_value.groupby(level=0).sum()
        quote_asset_value = quote_asset_value.groupby(level=0).sum()

        # get position
        position = pd.Series({i['asset']: i['free'] for i in self.ticker_info.info['balances']
                              if float(i['free']) != 0}).astype(float)
        position = position[position.index.str[:2] != 'LD']

        # refine asset index
        all_assets = base_asset_value.index.union(quote_asset_value.index).union(position.index)

        base_asset_value = base_asset_value.reindex(all_assets).fillna(0)
        quote_asset_value = quote_asset_value.reindex(all_assets).fillna(0)
        position = position.reindex(all_assets).fillna(0)

        # calculate algo value
        algo_value_in_btc = base_asset_value + quote_asset_value
        asset_price_in_btc = position.index.map(self.ticker_info.get_asset_price_in_btc)
        algo_value = algo_value_in_btc / asset_price_in_btc

        # calculate diffierence
        margin_position = pd.Series(self._margins).reindex(all_assets).fillna(0)

        diff_value_btc = pd.DataFrame({
            'algo_p': algo_value_in_btc,
            'margin_p': margin_position * asset_price_in_btc,
            'estimate_p': algo_value_in_btc + margin_position * asset_price_in_btc,
            'present_p': position * asset_price_in_btc,
            'difference': (algo_value_in_btc + margin_position * asset_price_in_btc).clip(0, None) - position * asset_price_in_btc,
            'rebalance_threshold': (algo_value_in_btc + margin_position * asset_price_in_btc).abs() * rebalance_threshold,
        })
        diff_value_btc['rebalance'] = diff_value_btc['difference'].abs() > diff_value_btc['rebalance_threshold']
        diff_value_btc.loc[quote_asset_list, 'rebalance'] = True

        # excluding checking of asset positions

        excluded = pd.Series(True, diff_value_btc.index)
        excluded[diff_value_btc.index.isin(signals.quote_asset) | diff_value_btc.index.isin(signals.base_asset)] = False
        excluded[diff_value_btc.index.isin(excluded_assets)] = True

        diff_value_btc['excluded'] = excluded

        diff_value = diff_value_btc.copy()
        diff_value = diff_value.div(asset_price_in_btc, axis=0)
        diff_value.rebalance = diff_value.rebalance != 0
        diff_value.excluded = diff_value.excluded != 0

        # calculate transactions

        rebalance_value_btc = diff_value_btc.rebalance * diff_value_btc.difference * (~diff_value_btc.excluded)
        increase_asset_amount = rebalance_value_btc[rebalance_value_btc > 0]
        decrease_asset_amount = rebalance_value_btc[rebalance_value_btc < 0]

        diff_value_btc['rebalance'] = diff_value_btc['difference'].abs() > diff_value_btc['rebalance_threshold']
        diff_value['rebalance'] = diff_value_btc.rebalance

        txn_btc = {}

        for nai, ai in increase_asset_amount.items():
            for nad, ad in decrease_asset_amount.items():

                symbol = nad + nai
                amount = min(-ad, ai)

                is_valid = self.ticker_info._list_select(self.ticker_info.tickers, 'symbol',
                                                         symbol) is not None and nai in quote_asset_list

                if is_valid:
                    increase_asset_amount.loc[nai] -= amount
                    decrease_asset_amount.loc[nad] += amount
                    txn_btc[symbol] = -amount
                    continue

                symbol = nai + nad
                is_valid = self.ticker_info._list_select(self.ticker_info.tickers, 'symbol',
                                                         symbol) is not None and nad in quote_asset_list

                if is_valid:
                    increase_asset_amount.loc[nai] -= amount
                    decrease_asset_amount.loc[nad] += amount
                    txn_btc[symbol] = amount
                    continue

        # assumption: self.default_stable_coin can be the quote asset for all alt-coins
        transaction_btc = increase_asset_amount.append(decrease_asset_amount)
        transaction_btc.index = transaction_btc.index + self.default_stable_coin

        if self.default_stable_coin in transaction_btc.index:
            transaction_btc.pop(self.default_stable_coin+self.default_stable_coin)

        transaction_btc = transaction_btc.append(pd.Series(txn_btc))

        transaction = transaction_btc.to_frame(name='value_in_btc')

        transaction['base_asset'] = transaction.index.map(self.ticker_info.get_base_asset)
        transaction['quote_asset'] = transaction.index.map(self.ticker_info.get_quote_asset)
        transaction['value'] = transaction['value_in_btc'] / transaction.base_asset.map(
            self.ticker_info.get_asset_price_in_btc)
        transaction['price'] = transaction.index.map(
            lambda s: self.ticker_info._list_select(self.ticker_info.tickers, 'symbol', s)['price'])
        transaction = transaction.groupby(level=0).agg(
            dict(value_in_btc='sum', value='sum', base_asset='first', quote_asset='first', price='first'))
        transaction = transaction[transaction.value != 0]

        # check difference after transaction
        def asset_distributed(v):
            asset_increase = v.value_in_btc.groupby(v.base_asset).sum()
            asset_decrease = v.value_in_btc.groupby(v.quote_asset).sum()
            return asset_increase.reindex(all_assets).fillna(0) - asset_decrease.reindex(all_assets).fillna(0)

        verify_assets = asset_distributed(transaction)
        verify_assets = verify_assets[verify_assets != 0]
        verify = (verify_assets / diff_value_btc.difference.reindex(verify_assets.index) - 1).abs() < 0.001

        try:
            assert verify[verify.index != self.default_stable_coin].all()
        except:
            print(diff_value_btc)
            print(transaction)
            print(verify_assets)
            print(verify)
            raise Exception("validation fail")

        # filter out orders where base asset is in quote asset list (ex: btcusdt)
        # assumption: base asset should only be paired by one quote asset
        transaction = transaction[
                        ~(transaction.base_asset.isin(quote_asset_list) &
                        (transaction.value_in_btc.abs() < diff_value_btc.loc[transaction['base_asset']].rebalance_threshold.values))
                        ]

        # verify diff_value
        def get_filters(exinfo, symbol):
            filters = self.ticker_info._list_select(self.ticker_info.exinfo['symbols'], 'symbol', symbol)['filters']
            min_lot_size = self.ticker_info._list_select(filters, 'filterType', 'LOT_SIZE')['minQty']
            step_size = self.ticker_info._list_select(filters, 'filterType', 'LOT_SIZE')['stepSize']
            min_notional = self.ticker_info._list_select(filters, 'filterType', 'MIN_NOTIONAL')['minNotional']
            return {
                'min_lot_size': min_lot_size,
                'step_size': step_size,
                'min_notional': min_notional,
            }

        filters = pd.DataFrame(
            {s: get_filters(self.ticker_info.exinfo, s) for s in transaction.index}).transpose().astype(float)

        if len(transaction) != 0:
            min_notional = filters.min_notional
            minimum_lot_size = filters.min_lot_size
            step_size = filters.step_size

            # rebalance filter:
            diff = transaction['value']

            # step size filter
            diff = round((diff / step_size).astype(int) * step_size, 9)

            # minimum lot filter
            diff[diff.abs() < minimum_lot_size] = 0

            # minimum notional filter
            diff[diff.abs() * transaction.price.astype(float) < min_notional] = 0

            transaction['final_value'] = diff
            transaction['final_value_in_btc'] = diff * transaction.base_asset.map(
                self.ticker_info.get_asset_price_in_btc)
        else:
            transaction = pd.DataFrame(None, columns=['final_value'])

        transaction = transaction[transaction['final_value'] != 0]

        return diff_value, diff_value_btc, transaction

    def execute_orders(self, transactions, mode='TEST'):
        """Execute orders to Binance.

        Execute orders by program order.

        Args:
          transactions: A dataframe which is generated by transaction in calculate_position_size() function result.
          mode: A str of transactions mode, we have 3 method.
              'TEST' is simulation.
              'MARKET' is market order which is transaction at the current latest price.
              'LIMIT' is The transaction is done at the specified price. If the specified price is not touched,
                      the transaction has not been completed.

        Returns:
            A dataframe of trades.
        """
        def cancel_orders(symbol):
            orders = self._client.get_open_orders(symbol=symbol)
            for o in orders:
                self._client.cancel_order(symbol=symbol, orderId=o['orderId'])

        order_func = self._client.create_order if mode == 'MARKET' or mode == 'LIMIT' else self._client.create_test_order

        print('|---------EXECUTION LOG----------|')
        print('| time: ', datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'))

        trades = {}
        for s, lot in transactions.final_value.items():

            cancel_orders(s)

            if lot == 0:
                continue

            side = SIDE_BUY if lot > 0 else SIDE_SELL
            try:
                args = dict(
                    side=side,
                    type=ORDER_TYPE_MARKET,
                    symbol=s,
                    quantity=abs(lot))

                if mode == 'LIMIT':
                    args['price'] = transactions.price.loc[s]
                    args['type'] = ORDER_TYPE_LIMIT
                    args['timeInForce'] = 'GTC'

                order_func(**args)
                order_result = 'success'
                print('|', mode, s, side, abs(lot), order_result)
            except Exception as e:
                print('| FAIL', s, s, side, abs(lot), str(e))
                order_result = 'FAIL: ' + str(e)

            trades[s] = {
                **args,
                'result': order_result,
            }

        return pd.DataFrame(trades).transpose()

    def status(self, ohlcvs):
        """Strategy list widgets.

        Choose which strategy to implement on widgets GUI.

        Args:
          ohlcvs: A dataframe of symbol.

        Returns:
            widget GUI
        """
        import ipywidgets as widgets

        ret = pd.DataFrame()
        full_results = []
        for method in self._trading_methods:
            for symbol in method.symbols:
                ohlcv = ohlcvs[(symbol, method.freq)]
                result = method.strategy.backtest(ohlcv,
                                                  method.variables, filters=method.filters, freq=method.freq)
                ret[method.name + '-' + symbol + '-' + method.freq] = result.cumulative_returns

                weight_btc = method.weight_btc
                if isinstance(weight_btc, dict):
                    weight_btc = (weight_btc[symbol] if symbol in weight_btc else weight_btc['default'])

                full_results.append({
                    'name': method.name,
                    'symbol': symbol,
                    'freq': method.freq,
                    'weight': weight_btc,
                    'portfolio': result,
                    'trading_method': method,
                    'signal': result.cash().iloc[-1] == 0,
                })

        method_dropdown = widgets.Dropdown(options=[m.name + '-' + str(i) for i, m in enumerate(self._trading_methods)])
        symbol_dropdown = widgets.Dropdown(options=[symbol + '-' + freq for symbol, freq in ohlcvs.keys()])
        backtest_btn = widgets.Button(description='status')

        backtest_panel = widgets.Output()
        option_panel = widgets.Output()

        def plotly_df(df):
            """Display plot.
            """
            # Plot
            fig = px.line()
            for sname, s in df.items():
                fig.add_scatter(x=s.index, y=s.values, name=sname)  # Not what is desired - need a line
            # fig.show()

        @backtest_panel.capture(clear_output=True)
        def backtest(_):
            """Display single strategy backtest result.
            """
            method_id = int(method_dropdown.value.split('-')[-1])
            history_id = tuple(symbol_dropdown.value.split('-'))
            ohlcv = ohlcvs[history_id]
            strategy = self._trading_methods[method_id].strategy
            svars = self._trading_methods[method_id].variables
            filters = self._trading_methods[method_id].filters
            strategy.backtest(ohlcv, variables=svars, filters=filters, freq=history_id[-1], plot=True)

        backtest_btn.on_click(backtest)
        dropdowns = widgets.HBox([method_dropdown, symbol_dropdown, backtest_btn])
        with option_panel:
            plotly_df(ret)
            display(pd.DataFrame(full_results))

        return widgets.VBox([option_panel, dropdowns, backtest_panel])

    def portfolio_backtest(self, ohlcvs, min_freq, quote_assets=['BTC', 'USDT', 'BUSD', 'USDC'], fee=0.002, delay=0):
        """Display portfolio backtest result.

        Calculate overall account asset changes.
        Unit is USD

        Args:
          ohlcvs: A dataframe of symbel.
          min_freq: A str of calculation frequency ex('4h').
          quote_assets: A list of assets name ex(['BTC', 'USDT', 'BUSD', 'ETH']).
          fee: A float of trading fee.
          delay: A int of delayed entry and exit setting.
        Returns:
            widget GUI
        """
        # backtest_results
        results = []
        for method in self._trading_methods:
            for symbol in method.symbols:
                ohlcv = ohlcvs[(symbol, method.freq)]
                result = method.strategy.backtest(ohlcv,
                                                  method.variables, filters=method.filters, freq=method.freq)
                # find weight_btc if it is in the nested dictionary
                weight_btc = method.weight
                if isinstance(weight_btc, dict):
                    weight_btc = (weight_btc[symbol] if symbol in weight else weight_btc['default'])

                weight_btc *= self.ticker_info.get_asset_price_in_btc(method.weight_unit)

                results.append({
                    'name': method.name,
                    'symbol': symbol,
                    'freq': method.freq,
                    'weight': weight_btc,
                    'portfolio': result,
                    'trading_method': method,
                    'signal': result.cash().iloc[-1] == 0,
                })

        results = pd.DataFrame(results)

        import matplotlib.pyplot as plt
        position = {}
        # quote_substract = {}
        for index, value in results.transpose().items():
            position[value.loc['name'] + '|' + value.symbol + '|' + value.freq] = (value.portfolio.cash() == 0).shift(
                delay).ffill() * value.weight
        position = pd.DataFrame(position).resample(min_freq).last().ffill()
        position.columns = position.columns.str.split('|').str[1]
        position = position.ffill().fillna(0)
        position = position.groupby(position.columns, axis=1).sum()

        # find quote assets
        quote_asset_col = []
        for symbol in position.columns:
            for q in quote_assets:
                if symbol[-len(q):] == q:
                    quote_asset_col.append(q)
                    break

        quote_position = position.copy()
        quote_position.columns = quote_asset_col
        quote_position = -quote_position.groupby(quote_position.columns, axis=1).sum()

        # calculate return in usdt
        assets = position.columns.str.split('|').str[0].to_list()

        for i, a in enumerate(assets):
            for q in quote_assets:
                if len(a) > 5 and a[-len(q):] == q:
                    assets[i] = a[:-len(q)]

        position.columns = assets
        position = position.groupby(position.columns, axis=1).sum()
        quote_position = quote_position.groupby(quote_position.columns, axis=1).sum()

        all_symbols = list(set(quote_position.columns) | set(position.columns) | set(self._margins.keys()))
        if 'USDT' not in all_symbols:
            all_symbols.append('USDT')

        position = position.reindex(all_symbols, axis=1).fillna(0) + quote_position.reindex(all_symbols, axis=1).fillna(
            0)

        ohlcv_usdt = {a: get_all_binance(a + 'USDT', min_freq) for a in position.columns if a != 'USDT'}

        initial_margin_sum_btc = 0
        for a, w in self._margins.items():
            position[a] += self.ticker_info.get_asset_price_in_btc(a) * w
            initial_margin_sum_btc += self.ticker_info.get_asset_price_in_btc(a) * w

        # remove negative position
        negative_position = ((position < 0) * position).drop('USDT', axis=1, errors='ignore').sum(axis=1)
        pusdt = position['USDT'].copy()
        position = position.clip(0, None)
        position.USDT = pusdt + negative_position

        addition_usdt = -min(position.USDT.min(), 0) / self.ticker_info.get_asset_price_in_btc('USDT')

        if addition_usdt > 0:
            print('WARRN**: additional usdt is required: ', addition_usdt, ' USD')

        p = position.loc[position.index[(position != position.shift()).abs().sum(axis=1) != 0] | position.index[-1:]]
        p.index = p.index.tz_localize(None)

        ohlcv_usdt_close = pd.DataFrame({name: s.close for name, s in ohlcv_usdt.items()})
        ohlcv_usdt_close.index = ohlcv_usdt_close.index.tz_localize(None)

        rebalance_time = (p.index & ohlcv_usdt_close.index)

        ohlcv_usdt_close = ohlcv_usdt_close.loc[rebalance_time]
        p = p.loc[rebalance_time].fillna(0)

        asset_return = ((ohlcv_usdt_close.pct_change().shift(-1).fillna(0)) * p) - fee * (p - p.shift()).abs()
        asset_return.fillna(0, inplace=True)

        (asset_return.cumsum() / self.ticker_info.get_asset_price_in_btc('USDT')).plot()
        plt.show()

        s = (asset_return.sum(axis=1).cumsum() + initial_margin_sum_btc) / self.ticker_info.get_asset_price_in_btc(
            'USDT')
        s.plot()

        plt.show()

        (s / s.cummax()).plot()
        plt.show()

        return results


def render_html(signals, position, position_btc, orders, order_results):
    """Render html to google cloud platform.

    Integrate order data into tables that display in html.

    Args:
      signals_df: A dataframe of signals which are generated by TradingPortfolio().get_latest_signals().
      rebalance_df: A dataframe of diff_value which are generated by TradingPortfolio().calculate_position_size().
      rebalance_df_in_btc: A dataframe of diff_value_btc which are generated by TradingPortfolio().calculate_position_size().
      orders: A dataframe of transaction which are generated by TradingPortfolio().calculate_position_size().
      order_results: A dataframe of execute_orders which are generated by TradingPortfolio().execute_orders().

    Returns:
        html
    """
    html = """
    <!DOCTYPE html>
    <head>
      <title>Saying Hello</title>
      <link rel="stylesheet" href="https://unpkg.com/purecss@2.0.3/build/pure-min.css"
      integrity="sha384-cg6SkqEOCV1NbJoCu11+bm0NvBRc8IYLRGXkmNrqUBfTjmMYwNKPWBTIKyw9mHNJ" crossorigin="anonymous">
      <meta name="viewport" content="width=device-width, initial-scale=1">

    </head>
    <body style="padding: 5vw">
    """

    html += '<h1>Crypto Portfolio</h1>'
    html += '<h2>Strategy signals</h2>'
    html += signals.to_html(classes="pure-table pure-table-horizontal")
    html += '<h2>Position</h2>'
    html += position.to_html(classes="pure-table pure-table-horizontal")
    html += '<h2>Position in BTC</h2>'
    html += position_btc.to_html(classes="pure-table pure-table-horizontal")
    html += '<h2>Orders</h2>'

    if len(orders) > 0:
        orders['result'] = order_results['result']
        html += orders.to_html(classes="pure-table pure-table-horizontal")
    else:
        html += '<p>None</p>'

    html += '<br>'
    html += '<button onclick="update_position(\'MARKET\')">place market orders</button>'
    html += '<button onclick="update_position(\'LIMIT\')">place limit orders</button>'
    html += '</body>'

    html += """
  <script>
  function update_position(mode) {
    // Redirect to next page
    var next_page = window.location.href.split("?")[0] + "?mode=" + mode
    window.location = next_page;
  }
  </script>
  """
    return html



import numpy as np
import pandas as pd
from typing import Union
from finlab_crypto.indicators import trends
from crypto_strategy.data import (
    save_stats,
    get_acc_returns,
    download_crypto_history
)
from .base import (
    trend_strategy,
    mmi_filter, ang_filter,
    BestStrategy, InspectStrategy, CheckIndicators
) 


RANGE_WINDOW = np.arange(30, 300, 10)
RANGE_TIMEPERIOD = np.arange(10, 50, 5)
RANGE_THRESHOLD = np.arange(0, 25, 5)
RANGE_STOP = np.arange(0.1, 0.6, 0.05)


def create_mmi_filter(**kwargs):
    if kwargs:
        assert kwargs.get('timeperiod'),\
            'mmi fitler doesn\'t have config params provided'
        filter_config = kwargs
    else:
        filter_config = dict(
            timeperiod=RANGE_TIMEPERIOD,
        )
    filters = {
        'mmi': mmi_filter.create({
            'timeperiod': filter_config['timeperiod']
        })
    }
    return filters


def create_ang_filter(**kwargs):
    if kwargs:
        assert kwargs.get('timeperiod') and kwargs.get('threshold') is not None,\
            'Angle filter doesn\'t have config params provided'
        filter_config = dict(
            timeperiod=kwargs['timeperiod'],
            threshold=kwargs['threshold']
        )
    else:
        filter_config = dict(
            timeperiod=RANGE_TIMEPERIOD,
            threshold=RANGE_THRESHOLD
        )
    filters = {
        'ang': ang_filter.create({
            'timeperiod': filter_config['timeperiod'],
            'threshold': filter_config['threshold']
        })
    }
    return filters


def get_filter(flag_filter, **kwargs):
    filters = dict()
    if flag_filter == 'mmi':
        filters = create_mmi_filter(**kwargs)
    if flag_filter == 'ang':
        filters = create_ang_filter(**kwargs)
    return filters


def create_ma_variables(**kwargs):
    if kwargs.get('name'):
        variables = dict(name=kwargs.get('name'))
    else:
        raise ValueError('The name of the MA strategy is required')
    # Load given params
    if kwargs.get('n1') and kwargs.get('n2'):
        variables.update(kwargs)
    # Set up params for backtests
    elif not (kwargs.get('n1') or kwargs.get('n2')):
        variables['n1'] = RANGE_WINDOW
        variables['n2'] = RANGE_WINDOW
    else:
        raise ValueError('The MA strategy does\'t have the config params provided')
    return variables


def create_ma_variables_with_stop(**kwargs):
    variables = create_ma_variables(**kwargs)
    if not (kwargs.get('n1') or kwargs.get('n2')):
        flag_stop = kwargs.get('flag_stop')
        if kwargs.get('flag_stop'):
            if not set(flag_stop).issubset(['ts_stop', 'sl_stop', 'tp_stop']):
                raise ValueError('The value of flag_stop is not supported')
            for fs in flag_stop:
                variables[fs] = RANGE_STOP
        else:
            raise ValueError('flag_stop is not set')
    return variables


class BestMaStrategy(BestStrategy):
    '''
    This class provides the method to optimize the MA strategy
    symbols: a single symbol or a list of symbols to be optimzied on, e.g., 'BTCUSDT'
    freq: currently supported values are '1h' or '4h'
    res_dir: the output directory
    flag_filter: flag to specify filters, currently support 'mmi', 'ang', default None
    flag_stop: flag to specify early stop, currently support 'ts_stop', 'sl_stop', 'tp_stop', default None
    trends: a list of MA strategies, default: trends.keys()
    strategy: strategy name, default: 'ma'
    '''
    def __init__(self,
                 symbols: Union[str, list],
                 freq: str,
                 res_dir: str,
                 flag_filter: str = None,
                 flag_stop: Union[str, list] = None,
                 flag_acc_return: bool = True,
                 trends: list = trends.keys(),
                 strategy: str = 'ma'
                 ):
        super().__init__(symbols, freq, res_dir, flag_filter, strategy)
        self.trends = trends
        self.flag_stop = [flag_stop] if isinstance(flag_stop, str) else flag_stop
        self.flag_acc_return = flag_acc_return
        self.generate_best_params()

    def _get_strategy(self, strategy):
        return trend_strategy

    def _get_filter(self, **kwargs):
        return get_filter(flag_filter=self.flag_filter, **kwargs)

    def _get_variables(self, **kwargs):
        if self.flag_stop:
            return create_ma_variables_with_stop(flag_stop=self.flag_stop, **kwargs)
        return create_ma_variables(**kwargs)

    def _get_grid_search(self):
        if self.flag_filter == 'mmi':
            return self.grid_search_mmi_params()
        if self.flag_filter == 'ang':
            return self.grid_search_ang_params()
        return self.grid_search_params()

    def get_best_params(self, total_best_params, n=10):
        total_best_params = pd.DataFrame(total_best_params)
        total_best_params['sharpe'] = (total_best_params['sharpe'] + 0.05).round(1)
        total_best_params['gap'] = total_best_params['n2'] - total_best_params['n1']
        # TODO: add tp_stop in sort_values
        if 'ohlcstx_sl_stop' in total_best_params.columns:
            total_best_params = (
                total_best_params
                .query('gap > 0')
                .sort_values(
                    by=['sharpe', 'gap', 'n1', 'ohlcstx_sl_stop'],
                    ascending=[True, True, False, False])
            )
        else:
            total_best_params = (
                total_best_params
                .query('gap > 0')
                .sort_values(
                    by=['sharpe', 'gap', 'n1'],
                    ascending=[True, True, False]
                    )
            )
        print(total_best_params)
        return total_best_params.tail(1).to_dict(orient='records')[0] if not total_best_params.empty else None

    def grid_search_params(self):
        variables = self._get_variables(name=self.trends)
        best_params = self.backtest(variables)
        return self.get_best_params(best_params)

    def grid_search_mmi_params(self):
        variables = self._get_variables(name=self.trends)
        total_best_params = list()
        for timeperiod in RANGE_TIMEPERIOD:
            filters = self._get_filter(timeperiod=timeperiod)
            best_params = self.backtest(variables, filters)
            total_best_params.extend(best_params)
        return self.get_best_params(total_best_params)

    def grid_search_ang_params(self):
        variables = self._get_variables(name=self.trends)
        total_best_params = list()
        for threshold in RANGE_THRESHOLD:
            for timeperiod in RANGE_TIMEPERIOD:
                filters = self._get_filter(timeperiod=timeperiod, threshold=threshold)
                best_params = self.backtest(variables, filters)
                total_best_params.extend(best_params)
        return self.get_best_params(total_best_params)

    def apply_best_params(self, best_params, symbol):
        variables = self._get_variables(
            name=best_params['name'],
            n1=best_params['n1'],
            n2=best_params['n2']
        )
        if 'ts_stop' in self.flag_stop:
            variables.update(ts_stop=best_params.get('ohlcstx_sl_stop'))
        if 'sl_stop' in self.flag_stop:
            variables.update(sl_stop=best_params.get('ohlcstx_sl_stop'))
        if 'tp_stop' in self.flag_stop:
            variables.update(tp_stop=best_params.get('ohlcstx_tp_stop'))
        filters = self._get_filter(
            timeperiod=best_params.get('mmi_timeperiod') or best_params.get('ang_timeperiod'),
            threshold=best_params.get('ang_threshold')
            )
        filename = f"{symbol}-{self.freq}-{best_params['name']}-{best_params['n1']}-{best_params['n2']}-"
        if self.flag_stop:
            if 'sl_stop' in self.flag_stop:
                filename += f"sl_stop-{best_params.get('ohlcstx_sl_stop'):.2f}-"
            if 'ts_stop' in self.flag_stop:
                filename += f"ts_stop-{best_params.get('ohlcstx_sl_stop'):.2f}-"
            if 'tp_stop' in self.flag_stop:
                filename += f"tp_stop-{best_params.get('ohlcstx_tp_stop'):.2f}-"
        if self.flag_filter == 'mmi':
            filename += f"{self.flag_filter}-{best_params['mmi_timeperiod']}-{self.date_str}.pkl"
        elif self.flag_filter == 'ang':
            filename += f"{self.flag_filter}-{best_params['ang_timeperiod']}-{best_params['ang_threshold']}-{self.date_str}.pkl"
        else:
            filename += f"{self.date_str}.pkl"
        portfolio = self.strategy.backtest(self.ohlcv, variables=variables, filters=filters, freq=self.freq)
        stats = portfolio.stats()
        if self.flag_acc_return:
            acc_returns = get_acc_returns(portfolio.daily_returns())
            stats = stats.append(pd.Series(acc_returns))
        save_stats(stats, self.output_path, filename)
        print(f'The stats are saved to {self.output_path}/{filename}')

    def generate_best_params(self):
        print(f'The search for {self.symbols} starts now')
        for symbol in self.symbols:
            print(f'Start the search for {symbol}')
            self.ohlcv = download_crypto_history(symbol, self.freq)
            best_params = self._get_grid_search()
            print(f'Found the best params {best_params}')
            self.apply_best_params(best_params, symbol)
        print('The search for all the symbols is completed')


class CheckMaIndicators(CheckIndicators):
    '''
    This class provides the method to check Partial Differentiation
    and Combinatorially Symmetric Cross-validation of MA strategy.
    symbols: a list of symbols to be optimzied on, e.g., ['BTCUSDT']
    date: the date the best params are created
    res_dir: the output directory
    flag_filter: currently supported fitlers: 'mmi', 'ang', default: None
    flag_stop: flag to specify early stop, currently support 'ts_stop', 'sl_stop', 'tp_stop', default None
    '''
    def __init__(self,
                 symbols: Union[str, list],
                 date: str,
                 res_dir: str,
                 flag_filter: str = None,
                 flag_stop: Union[str, list] = None,
                 strategy: str = 'ma'
                 ):
        super().__init__(symbols, date, res_dir, flag_filter, strategy)
        self.flag_stop = [flag_stop] if isinstance(flag_stop, str) else flag_stop
        self.check_indicators()

    def _get_strategy(self, strategy):
        return trend_strategy

    def _get_variables(self, **kwargs):
        if self.flag_stop:
            return create_ma_variables_with_stop(flag_stop=self.flag_stop, **kwargs)
        return create_ma_variables(**kwargs)

    def _get_filter(self, **kwargs):
        return get_filter(flag_filter=self.flag_filter, **kwargs)


# TODO: dictionary for filters
class InspectMaStrategy(InspectStrategy):
    '''
    This class provides a method to inspect the MA strategy with given params
    symbol: the name of the crypto, e.g., 'BTCUSDT'
    freq: currently supported values are '1h' or '4h'
    name, n1, n2: the name and the params of the ma strategy, e.g., 'sma', 100, 50
    timeperiod: param used in either mmi or ang filter
    threshold: param used in ang filter
    flag_filter: currently supported fitlers: 'mmi', 'ang', default: None
    stop_vars: dictionary of stop vars, currently support 'ts_stop', 'sl_stop', 'tp_stop', default None
    '''
    def __init__(self,
                 symbol: str,
                 freq: str,
                 name: str,
                 n1: int,
                 n2: int,
                 flag_filter: str = None,
                 timeperiod: int = None,
                 threshold: int = None,
                 stop_vars: dict = None,
                 strategy: str = 'ma',
                 show_fig: bool = True
                 ):
        super().__init__(symbol, freq, flag_filter, strategy, show_fig)
        self.name = name
        self.n1 = n1
        self.n2 = n2
        self.timeperiod = timeperiod
        self.threshold = threshold
        self.flag_filter = flag_filter
        self.stop_vars = stop_vars
        self.inspect()

    def _get_strategy(self, strategy):
        return trend_strategy

    def _get_variables(self):
        variables = create_ma_variables(name=self.name, n1=self.n1, n2=self.n2)
        if self.stop_vars:
            variables.update(self.stop_vars)
        return variables

    def _get_filter(self):
        return get_filter(
            self.flag_filter,
            timeperiod=self.timeperiod,
            threshold=self.threshold
            )


def returns_timeline(
    symbol: str,
    freq: str,
    name: str,
    n1: int,
    n2: int,
    flag_filter: str = None,
    timeperiod: int = None,
    threshold: int = None,
    stop_vars: dict = None,
    strategy: str = 'ma'
):
    ins = InspectMaStrategy(
        symbol=symbol,
        freq=freq,
        n1=n1,
        n2=n2,
        timeperiod=timeperiod,
        threshold=threshold,
        flag_filter=flag_filter,
        stop_vars = stop_vars,
        strategy=strategy,
        show_fig=False
    )
    daily_returns = ins.portfolio.daily_returns()
    acc_returns = get_acc_returns(daily_returns) 
    return acc_returns


import numpy as np
import pandas as pd
from crypto_strategy.data import download_crypto_history, save_stats
from .base import (
    macd_strategy, macd_strategy_revised,
    mmi_filter, ang_filter, stoch_filter, sma_filter,
    BestStrategy, InspectStrategy, CheckIndicators
)

RANGE_FASTPERIOD = np.arange(10, 200, 10)
RANGE_SLOWPERIOD = np.arange(10, 350, 10)
RANGE_SIGNALPERIOD = np.arange(10, 350, 10)
RANGE_TIMEPERIOD = np.arange(10, 50, 5)
RANGE_THRESHOLD = np.arange(0, 25, 5)
RANGE_FAST = np.arange(5, 15, 5)
RANGE_SLOW = np.arange(3, 15, 3)


def get_strategy(strategy):
    if strategy == 'macd':
        return macd_strategy
    if strategy == 'macd_rev':
        return macd_strategy_revised
    raise ValueError('Strategy not recognized. Please choose from macd and macd_rev')


def create_mmi_filter(**kwargs):
    if kwargs:
        assert kwargs.get('timeperiod'),\
            'mmi fitler doesn\'t have config params provided'
        filter_config = kwargs
    else:
        filter_config = dict(
            timeperiod=RANGE_TIMEPERIOD,
        )
    filters = {
        'mmi': mmi_filter.create({
            'timeperiod': filter_config['timeperiod']
        })
    }
    return filters


def create_ang_filter(**kwargs):
    if kwargs:
        assert kwargs.get('timeperiod') and kwargs.get('threshold') is not None,\
            'Angle filter doesn\'t have config params provided'
        filter_config = dict(
            timeperiod=kwargs['timeperiod'],
            threshold=kwargs['threshold']
        )
    else:
        filter_config = dict(
            timeperiod=RANGE_TIMEPERIOD,
            threshold=RANGE_THRESHOLD
        )
    filters = {
        'ang': ang_filter.create({
            'timeperiod': filter_config['timeperiod'],
            'threshold': filter_config['threshold']
        })
    }
    return filters


def create_stoch_filter(**kwargs):
    if kwargs:
        assert kwargs.get('fast') and kwargs.get('slow') is not None,\
            'Stoch filter doesn\'t have config params provided'
        filter_config = dict(
            fast=kwargs['fast'],
            slow=kwargs['slow']
        )
    else:
        filter_config = dict(
            fast=RANGE_FAST,
            slow=RANGE_SLOW
        )
    filters = {
        'stoch': stoch_filter.create({
            'fast': filter_config['fast'],
            'slow': filter_config['slow']
        })
    }
    return filters


def create_sma_filter(**kwargs):
    if kwargs:
        assert kwargs.get('timeperiod') is not None,\
            'SMA filter doesn\'t have config params provided'
        filter_config = dict(
            timeperiod=kwargs['timeperiod']
        )
    else:
        filter_config = dict(
            timeperiod=RANGE_TIMEPERIOD
        )
    filters = {
        'sma': sma_filter.create({
            'timeperiod': filter_config['timeperiod']
        })
    }
    return filters


def get_filter(flag_filter, **kwargs):
    filters = dict()
    if flag_filter == 'mmi':
        filters = create_mmi_filter(**kwargs)
    if flag_filter == 'ang':
        filters = create_ang_filter(**kwargs)
    if flag_filter == 'stoch':
        filters = create_stoch_filter(**kwargs)
    if flag_filter == 'sma':
        filters = create_sma_filter(**kwargs)
    return filters


def create_variables(**kwargs):
    variables = dict()
    if kwargs.get('fastperiod') and kwargs.get('slowperiod') and kwargs.get('signalperiod'):
        variables.update(kwargs)
    else:
        variables['fastperiod'] = RANGE_FASTPERIOD
        variables['slowperiod'] = RANGE_SLOWPERIOD
        variables['signalperiod'] = RANGE_SIGNALPERIOD
    return variables


class BestMacdStrategy(BestStrategy):
    '''
    This class provides the method to optimize the MA strategy
    symbols: a list of symbols to be optimzied on, e.g., ['BTCUSDT']
    freq: currently supported values are '1h' or '4h'
    res_dir: the output directory
    flag_fitler: currently supported fitlers: 'mmi', 'ang', 'stoch', 'sma', default: None
    trends: a list of MA strategies, default: trends.keys()
    strategy: currently supported strategies: 'macd', 'macd_rev', default: 'macd'
    '''
    def __init__(self, symbols: list, freq: str, res_dir: str,
                 flag_filter: str = None,
                 strategy: str = 'macd'
                 ):
        super().__init__(symbols, freq, res_dir, flag_filter, strategy)
        self.generate_best_params()

    def _get_strategy(self, strategy):
        return get_strategy(strategy)

    def _get_filter(self, **kwargs):
        return get_filter(flag_filter=self.flag_filter, **kwargs)

    def _get_variables(self, **kwargs):
        return create_variables(**kwargs)

    def get_best_params(self, total_best_params, n=10):
        total_best_params = (
            pd.DataFrame(total_best_params)
            .sort_values(by=['sharpe', 'slowperiod'], ascending=[True, True])
            .tail(n)
        )
        print(total_best_params)
        return total_best_params.tail(1).to_dict(orient='records')[0]

    def _get_grid_search(self):
        if self.flag_filter == 'mmi':
            return self.grid_search_mmi_params()
        if self.flag_filter == 'ang':
            return self.grid_search_ang_params()
        if self.flag_filter == 'stoch':
            return self.grid_search_stoch_params()
        if self.flag_filter == 'sma':
            return self.grid_search_sma_params()
        return self.grid_search_params()

    def grid_search_params(self):
        variables = self._get_variables()
        best_params = self.backtest(variables)
        return self.get_best_params(best_params)

    def grid_search_mmi_params(self):
        variables = self._get_variables()
        total_best_params = list()
        for timeperiod in RANGE_TIMEPERIOD:
            filters = self._get_filter(timeperiod=timeperiod)
            best_params = self.backtest(variables, filters)
            total_best_params.extend(best_params)
        return self.get_best_params(total_best_params)

    def grid_search_ang_params(self):
        variables = self._get_variables()
        total_best_params = list()
        for threshold in RANGE_THRESHOLD:
            for timeperiod in RANGE_TIMEPERIOD:
                filters = self._get_filter(timeperiod=timeperiod, threshold=threshold)
                best_params = self.backtest(variables, filters)
                total_best_params.extend(best_params)
        return self.get_best_params(total_best_params)

    def grid_search_stoch_params(self):
        variables = self._get_variables()
        total_best_params = list()
        for fast in RANGE_FAST:
            for slow in RANGE_SLOW:
                filters = self._get_filter(fast=fast, slow=slow)
                best_params = self.backtest(variables, filters)
                total_best_params.extend(best_params)
        return self.get_best_params(total_best_params)

    def grid_search_sma_params(self):
        variables = self._get_variables()
        total_best_params = list()
        for timeperiod in RANGE_TIMEPERIOD:
            filters = self._get_filter(timeperiod=timeperiod)
            best_params = self.backtest(variables, filters)
            total_best_params.extend(best_params)
        return self.get_best_params(total_best_params)

    def apply_best_params(self, best_params, symbol):
        variables = self._get_variables(
            fastperiod=best_params['fastperiod'],
            slowperiod=best_params['slowperiod'],
            signalperiod=best_params['signalperiod']
            )
        filters = self._get_filter(
            timeperiod=best_params.get('mmi_timeperiod')
            or best_params.get('ang_timeperiod')
            or best_params.get('sma_timeperiod'),
            threshold=best_params.get('ang_threshold'),
            fast=best_params.get('stoch_fast'),
            slow=best_params.get('stoch_slow')
            )
        portfolio = self.strategy.backtest(self.ohlcv, variables=variables, filters=filters, freq=self.freq)
        filename = f"{symbol}-{self.freq}-{best_params['fastperiod']}-{best_params['slowperiod']}-{best_params['signalperiod']}-"
        if self.flag_filter == 'mmi':
            filename += f"{self.flag_filter}-{best_params['mmi_timeperiod']}-{self.date_str}.pkl"
        elif self.flag_filter == 'ang':
            filename += f"{self.flag_filter}-{best_params['ang_timeperiod']}-{best_params['ang_threshold']}-{self.date_str}.pkl"
        elif self.flag_filter == 'stoch':
            filename += f"{self.flag_filter}-{best_params['stoch_fast']}-{best_params['stoch_slow']}-{self.date_str}.pkl"
        elif self.flag_filter == 'sma':
            filename += f"{self.flag_filter}-{best_params['sma_timeperiod']}-{self.date_str}.pkl"
        else:
            filename += f"{self.date_str}.pkl"
        save_stats(portfolio.stats(), self.output_path, filename)
        print(f'The stats are saved to {self.output_path}/{filename}')

    def generate_best_params(self):
        print(f'The search for {self.symbols} starts now')
        for symbol in self.symbols:
            print(f'Start the search for {symbol}')
            self.ohlcv = download_crypto_history(symbol, self.freq)
            best_params = self._get_grid_search()
            print(f'Found the best params {best_params}')
            self.apply_best_params(best_params, symbol)
        print('The search for all the symbols is completed')


class CheckMacdIndicators(CheckIndicators):
    '''
    This class provides the method to check Partial Differentiation
    and Combinatorially Symmetric Cross-validation of MACD strategy
    symbols: a list of symbols to be optimzied on, e.g., ['BTCUSDT']
    date: the date the best params are created
    res_dir: the output directory
    flag_fitler: currently supported fitlers: 'mmi', 'ang', 'stoch', 'sma', default: None
    strategy: MACD strategy
    '''
    def __init__(self, symbols: list, date: str, res_dir: str,
                 flag_filter: str = None, strategy: str = 'macd'
                 ):
        super().__init__(symbols, date, res_dir, flag_filter, strategy)
        self.check_indicators()

    def _get_strategy(self, strategy):
        return get_strategy(strategy)

    def _get_variables(self, **kwargs):
        return create_variables(**kwargs)

    def _get_filter(self, **kwargs):
        return get_filter(flag_filter=self.flag_filter, **kwargs)


class InspectMacdStrategy(InspectStrategy):
    '''
    This class provides a method to inspect the MA strategy with given params
    symbol: the name of the crypto, e.g., 'BTCUSDT'
    freq: currently supported values are '1h' or '4h'
    fastperiod, slowperiod, signalperiod: params of macd strategy
    timeperiod, threshold, fast, slow: params used in filters
    flag_fitler: currently supported fitlers: 'mmi', 'ang', 'stoch', 'sma', default: None
    '''
    def __init__(self,
                 symbol: str, freq: str,
                 fastperiod: int, slowperiod: int, signalperiod: int,
                 timeperiod: int = None, threshold: int = None,
                 fast: int = None, slow: int = None,
                 flag_filter: str = None,
                 strategy: str = 'macd',
                 show_fig: bool = True
                 ):
        super().__init__(symbol, freq, flag_filter, strategy, show_fig)
        self.fastperiod = fastperiod
        self.slowperiod = slowperiod
        self.signalperiod = signalperiod
        self.timeperiod = timeperiod
        self.threshold = threshold
        self.fast = fast
        self.slow = slow
        self.flag_filter = flag_filter
        self.inspect()

    def _get_strategy(self, strategy):
        return get_strategy(strategy)

    def _get_variables(self):
        return create_variables(
            fastperiod=self.fastperiod,
            slowperiod=self.slowperiod,
            signalperiod=self.signalperiod
            )

    def _get_filter(self):
        return get_filter(
            self.flag_filter,
            timeperiod=self.timeperiod,
            threshold=self.threshold,
            fast=self.fast,
            slow=self.slow
            )



import os
import pathlib
from typing import Union
from datetime import datetime
from abc import ABC, abstractmethod
import talib
import numpy as np
import pandas as pd
from finlab_crypto.indicators import trends
from finlab_crypto.strategy import Strategy, Filter
from crypto_strategy.reporting.plot import plot_indicators
from crypto_strategy.data import check_and_create_dir, download_crypto_history


@Strategy(name='sma', n1=20, n2=40)
def trend_strategy(ohlcv):
    name = trend_strategy.name
    n1 = trend_strategy.n1
    n2 = trend_strategy.n2
    filtered1 = trends[name](ohlcv.close, n1)
    filtered2 = trends[name](ohlcv.close, n2)
    entries = (filtered1 > filtered2) & (filtered1.shift() < filtered2.shift())
    exits = (filtered1 < filtered2) & (filtered1.shift() > filtered2.shift())
    figures = {
        'overlaps': {
            'trend1': filtered1,
            'trend2': filtered2,
        }
    }
    return entries, exits, figures


@Strategy(long_window=30, short_window=30)
def breakout_strategy(ohlcv):
    lw = breakout_strategy.long_window
    sw = breakout_strategy.short_window
    ub = ohlcv.close.rolling(lw).max()
    lb = ohlcv.close.rolling(sw).min()
    entries = ohlcv.close == ub
    exits = ohlcv.close == lb
    figures = {
        'overlaps': {
            'ub': ub,
            'lb': lb
        }
    }
    return entries, exits, figures


# macd strategy
@Strategy(fastperiod=12, slowperiod=26, signalperiod=9)
def macd_strategy(ohlcv):
    macd, signal, macdhist = talib.MACD(
        ohlcv.close,
        fastperiod=macd_strategy.fastperiod,
        slowperiod=macd_strategy.slowperiod,
        signalperiod=macd_strategy.signalperiod
    )

    entries = (macdhist > 0) & (macdhist.shift() < 0)
    exits = (macdhist < 0) & (macdhist.shift() > 0)
    figures = {
        'figures': {
            'macdhist': macdhist
        }
    }
    return entries, exits, figures


@Strategy(fastperiod=12, slowperiod=26, signalperiod=9)
def macd_strategy_revised(ohlcv):
    macd, signal, macdhist = talib.MACD(
        ohlcv.close,
        fastperiod=macd_strategy.fastperiod,
        slowperiod=macd_strategy.slowperiod,
        signalperiod=macd_strategy.signalperiod
    )
    entries = (macdhist > 0) & (macdhist.shift() < 0) & (macd > 0)
    exits = (macdhist < 0) & (macdhist.shift() > 0) & (macd < 0)
    figures = {
        'macd': {
            'macdhist': macdhist
            }
        }
    return entries, exits, figures


@Strategy(timeperiod=14, buy_threshold=52, sell_threshold=50)
def rsi_strategy(ohlcv):
    rsi = talib.RSI(ohlcv.close, timeperiod=rsi_strategy.timeperiod)
    entries = (rsi > rsi_strategy.buy_threshold) & (rsi.shift() < rsi_strategy.buy_threshold)
    exits = (rsi < rsi_strategy.sell_threshold) & (rsi.shift() > rsi_strategy.sell_threshold)
    figure = {
        'figures': {
            'rsi': rsi
        }
    }
    return entries, exits, figure


@Filter(timeperiod=20)
def mmi_filter(ohlcv):
    median = ohlcv.close.rolling(mmi_filter.timeperiod).median()
    p1 = ohlcv.close > median
    p2 = ohlcv.close.shift() > median
    mmi = (p1 & p2).astype(int).rolling(mmi_filter.timeperiod).mean()
    figures = {
      'figures': {
          'mmi_index': mmi
      }
    }
    return mmi >= 0.5, figures


@Filter(timeperiod=20, multiplier=2)
def vol_filter(ohlcv):
    vol_mean = ohlcv.volume.rolling(vol_filter.timeperiod).mean()
    vol = ohlcv.volume / (vol_filter.multiplier * vol_mean)
    figures = {
      'figures': {
          'vol_index': vol
      }
    }
    return vol >= 1, figures


@Filter(timeperiod=20, threshold=0)
def ang_filter(ohlcv):
    ang = talib.LINEARREG_ANGLE(ohlcv.close, ang_filter.timeperiod)
    figures = {
        'figures': {
            'ang_index': ang
        }
    }
    return ang > ang_filter.threshold, figures


@Filter(side='long', fast=5, slow=3, matype=0)
def stoch_filter(ohlcv):
    side = stoch_filter.side
    fast = stoch_filter.fast
    slow = stoch_filter.slow
    matype = stoch_filter.matype
    k, d = talib.STOCH(
        ohlcv.high, ohlcv.low, ohlcv.close,
        fastk_period=fast, slowk_period=slow, slowk_matype=matype,
        slowd_period=slow, slowd_matype=matype)
    signals = (k > d) & (k > 50) & (k < 50) if side == 'long' else k < d
    fig = {
        'figures': {
            'kd': {'k': k, 'd': d}
        }
    }
    return signals, fig


@Filter(timeperiod=50)
def sma_filter(ohlcv):
    sma = talib.SMA(ohlcv.close, timeperiod=sma_filter.timeperiod)
    figures = {
      'figures': {
          'sma': sma
      }
    }
    return ohlcv.close > sma, figures


class BestStrategy(ABC):
    def __init__(self,
                 symbols: Union[str, list],
                 freq: list,
                 res_dir: str,
                 flag_filter: str,
                 strategy: str):
        self._sanity_check(flag_filter, res_dir)
        self.symbols = [symbols] if isinstance(symbols, str) else symbols
        self.freq = freq
        self.flag_filter = flag_filter
        self.strategy_name = strategy
        self.strategy = self._get_strategy(strategy)
        self.date_str = datetime.today().date().strftime("%Y-%m-%d")
        self.output_path = check_and_create_dir(res_dir, self.date_str.replace('-', ''))
        self.output_path = pathlib.Path(self.output_path)

    @staticmethod
    def _sanity_check(flag_filter, res_dir):
        if flag_filter and flag_filter not in res_dir:
            raise ValueError(f'The res_dir name should contain {flag_filter} filter')
        if not flag_filter and 'filter' in res_dir:
            raise ValueError('The res_dir name should not contain filter name')

    @abstractmethod
    def _get_strategy(self, strategy):
        pass

    @abstractmethod
    def grid_search_params(self):
        pass

    @abstractmethod
    def _get_grid_search(self):
        pass

    @abstractmethod
    def apply_best_params(self):
        pass

    def get_best_params(self, total_best_params, n=10):
        total_best_params = (
            pd.DataFrame(total_best_params)
            .sort_values(by='sharpe')
        )
        print(total_best_params.head(n))
        return total_best_params.tail(1).to_dict(orient='records')[0]

    def backtest(self, variables, filters=dict(), n=10):
        portfolio = self.strategy.backtest(self.ohlcv, freq=self.freq, variables=variables, filters=filters)
        best_params = portfolio.sharpe_ratio().replace([np.inf, -np.inf], np.nan).dropna().nlargest(n)
        best_params = (
            best_params
            .rename('sharpe')
            .reset_index()
            .to_dict(orient='records')
        )
        return best_params

    def generate_best_params(self):
        print(f'The search for {self.symbols} starts now')
        for symbol in self.symbols:
            print(f'Start the search for {symbol}')
            self.ohlcv = download_crypto_history(symbol, self.freq)
            best_params = self._get_grid_search()
            print(f'Found the best params {best_params}')
            self.apply_best_params(best_params, symbol)
        print('The search for all the symbols is completed')


class CheckIndicators(ABC):
    def __init__(self,
                 symbols: Union[str, list],
                 date: str,
                 res_dir: str,
                 flag_filter: str,
                 strategy: str,
                 show_fig: bool = False):
        self.symbols = [symbols] if isinstance(symbols, str) else symbols
        self.date = date
        self.flag_filter = flag_filter
        self.res_dir = os.path.join(res_dir, date.replace('-', ''))
        self.strategy = self._get_strategy(strategy)
        self.show_fig = show_fig

    @abstractmethod
    def _get_strategy(self, strategy):
        pass

    @abstractmethod
    def _get_variables(self):
        pass

    @abstractmethod
    def _get_filter(self):
        pass

    def check_indicators(self):
        print(f'Start checking the optimized params for {self.symbols}')
        for symbol in self.symbols:
            if os.path.isdir(self.res_dir) and symbol:
                res = os.listdir(self.res_dir)
                for r in res:
                    if r.startswith(symbol) and r.endswith(self.date+'.pkl'):
                        comp = r.split('-')
                        assert symbol == comp[0], 'asset name doesn\'t match the file name'
                        freq = comp[1]
                        self.trend = comp[2] if comp[2].isalpha() else None
                        ohlcv = download_crypto_history(symbol, freq)
                        if self.trend:
                            variables = self._get_variables(name=self.trend)
                        else:
                            variables = self._get_variables()
                        filters = self._get_filter()
                        portfolio = self.strategy.backtest(ohlcv, freq=freq, variables=variables, filters=filters)
                        filename = f'{symbol}-{freq}-{self.trend}-{self.date}'
                        plot_indicators(portfolio, self.res_dir, filename, self.show_fig)


class InspectStrategy(ABC):
    def __init__(self,
                 symbol: str,
                 freq: str,
                 flag_filter: str,
                 strategy: str,
                 show_fig: bool = True):
        self.symbol = symbol
        self.freq = freq
        self.flag_filter = flag_filter
        self.strategy = self._get_strategy(strategy)
        self.show_fig = show_fig
        self.ohlcv = download_crypto_history(symbol, freq)

    @abstractmethod
    def _get_strategy(self, strategy):
        pass

    @abstractmethod
    def _get_variables(self):
        pass

    @abstractmethod
    def _get_filter(self):
        pass

    def inspect(self):
        self.portfolio = self.strategy.backtest(
            self.ohlcv,
            variables=self._get_variables(),
            filters=self._get_filter(),
            freq=self.freq,
            plot=self.show_fig)



import os
from collections import defaultdict
from datetime import datetime
import pandas as pd
from crypto_strategy.reporting.plot import plot_returns
from crypto_strategy.strategies.bo_strategy import returns_timeline
from .base import GenerateReport


class GenerateMaReport(GenerateReport):
    def __init__(self, symbols, date, res_dir):
        super().__init__(symbols, date, res_dir)

    def print_params(self, params):
        print(f'{params[0]:}')
        print(f'Best MA Strategy: {params[2]}')
        print(f'n1: {params[3]}')
        print(f'n2: {params[4]}')
        if 'mmi' in params:
            print(f'mmi_timeperiod: {params[6]}')


class GenerateBoReport(GenerateReport):
    def __init__(self, symbols, date, res_dir):
        super().__init__(symbols, date, res_dir)

    def print_params(self, params):
        print(f'{params[0]:}')
        print(f'long_window: {params[2]}')
        print(f'short_window: {params[3]}')
        if 'vol' in params:
            print(f'multiplier: {params[6]}')
            print(f'timeperiod: {params[5]}')


def generate_report(symbol, res_dir='results/'):
    metrics = [
        'Total Return [%]', 'Buy & Hold Return [%]',
        'Num. Trades', 'Win Rate [%]', 'Max. Drawdown [%]', 'Best Trade [%]', 'Worst Trade [%]',
        'Expectancy', 'Sharpe Ratio', 'Sortino Ratio', 'Calmar Ratio',
        'Start', 'End', 'Duration',
        'Ret [:21-04-14]', 'Ret [21-04-15:21-07-20]', 'Ret [21-07-21:21-11-10]', 'Ret [21-11-11:]',
        ]
    table = list()
    images = defaultdict(list)
    for tm_folder in os.listdir(res_dir):
        if tm_folder.startswith('best-'):
            cur_path = os.path.join(res_dir, tm_folder)
            date_folder = sorted([int(f) for f in os.listdir(cur_path) if not f.startswith('.')])
            asset_file = []
            _cur_path = cur_path
            # handle assets in multiple bases
            while not asset_file and date_folder:
                cur_path = os.path.join(_cur_path, str(date_folder[-1]))
                asset_file = [f for f in os.listdir(cur_path) if f.startswith(symbol)]
                date_folder.pop()
            for f in asset_file:
                entry = dict()
                if f.endswith('.pkl'):
                    entry['Method'] = tm_folder
                    entry['Params'] = '-'.join(f.split('-')[1:-3])
                    stats = pd.read_pickle(os.path.join(cur_path, f))
                    stats = stats[~stats.index.duplicated('last')]
                    for m in metrics:
                        # backwards compatible
                        if m not in stats:
                            if m == 'Buy & Hold Return [%]':
                                entry[m] = stats['Benchmark Return [%]']
                            if m == 'Num. Trades':
                                entry[m] = stats['Total Trades']
                            if m == 'Max. Drawdown [%]':
                                entry[m] = stats['Max Drawdown [%]']
                        else:
                            entry[m] = stats[m]
                    entry['Start'] = entry['Start'].strftime('%Y-%m-%d')
                    entry['End'] = entry['End'].strftime('%Y-%m-%d')
                    table.append(entry)
                if f.endswith('.png'):
                    images[tm_folder].append(os.path.join(cur_path, f))
    # create HTML file
    table = pd.DataFrame(table)
    table.sort_values(by='Method', inplace=True)
    date = datetime.today().strftime('%Y-%m-%d')
    heading = f'<h1> {symbol} Trading Strategies Testing Report</h1>'
    subheading = f'<h3> Best Params for each strategy {date}</h3>'
    header = f'<div class="top">{heading}{subheading}</div>'
    div_table = f'<div class="table"> {table.to_html()} </div>'
    div_images = ''
    for img in images:
        div_images += f'<h3> {img}</h3>'
        img_tag = f'<img src="../{images[img][0]}">'
        div_images += f'<div class="chart"> {img_tag} </div>'
        img_tag = f'<img src="../{images[img][1]}">'
        div_images += f'<div class="chart"> {img_tag} </div>'
    html_file = header + div_table + div_images
    with open(f'reports/report-{symbol}-{date}.html', 'w') as f:
        f.write(html_file)


def reporting_add_rets_timeline(symbol, method_path):
    if 'bo' in method_path.split('-'):
        strategy = 'bo'
    flag_ts_stop = True if 'ts_stop' in method_path else False
    if 'vol' in method_path:
        flag_filter = 'vol'
    elif 'ang' in method_path:
        flag_filter = 'ang'
    else:
        flag_filter = None
    date_folder = sorted([int(f) for f in os.listdir(method_path) if not f.startswith('.')])
    asset_file = []
    cur_path = method_path
    _cur_path = cur_path
    # handle assets in multiple bases
    while not asset_file and date_folder:
        cur_path = os.path.join(_cur_path, str(date_folder[-1]))
        asset_file = [f for f in os.listdir(cur_path) if f.startswith(symbol)]
        date_folder.pop()
    for f in asset_file:
        if f.endswith('.pkl'):
            print(f)
            parts = f.split('-')[:-3]
            symbol = parts[0]
            freq = parts[1]
            long_window = int(parts[3])
            short_window = int(parts[4])
            if flag_ts_stop:
                ts_stop = float(parts[6])
            else:
                ts_stop = None
            timeperiod, threshold, multiplier = None, None, None
            if flag_filter == 'ang':
                timeperiod = int(parts[-2])
                threshold = int(parts[-1])
            if flag_filter == 'vol':
                timeperiod = int(parts[-2])
                multiplier = int(parts[-1])
            ret = returns_timeline(
                symbol, freq,
                long_window, short_window,
                strategy, ts_stop,
                timeperiod, multiplier, threshold,
                flag_filter, flag_ts_stop
                )
            ser = pd.read_pickle(os.path.join(cur_path, f))
            if list(ret.keys())[-1] not in ser.index:
                ser = ser.append(pd.Series(ret))
                ser.to_pickle(os.path.join(cur_path, f))


def portfolio_returns(start_date, btc_ohlcv, eth_ohlcv, pf_daily_returns):
    if isinstance(start_date, str):
        start = start_date
        # print(f'Portfolio Returns Since {start_date}')
    if isinstance(start_date, int):
        # print(f'{start_date}-Day Portfolio Returns')
        start = -start_date
    # BTC | ETH Buy & Hold
    btc_daily_returns = btc_ohlcv[start:].close.pct_change().fillna(0)
    btc_acc_returns = (btc_daily_returns + 1).cumprod()
    # print(f'''BTC: {btc_acc_returns.last('1D') - 1}''')
    eth_daily_returns = eth_ohlcv[start:].close.pct_change().fillna(0)
    eth_acc_returns = (eth_daily_returns + 1).cumprod()
    # print(f'''ETH: {eth_acc_returns.last('1D') - 1}''')
    # Portfolio Returns
    pf_daily_returns = pf_daily_returns[start:].copy()
    pf_daily_returns.iloc[0] = 0
    pf_acc_returns = (pf_daily_returns + 1).cumprod().mean(axis=1)
    # print(f'''Portfolio: {pf_acc_returns.tail(1) - 1}''')
    plot_returns(btc_acc_returns, eth_acc_returns, pf_acc_returns, start_date)



from IPython.display import display, HTML
from itertools import product
from collections.abc import Iterable
import matplotlib.pyplot as plt
import tqdm.notebook as tqdm
import ipywidgets as widgets
import vectorbt as vbt
import seaborn as sns
import pandas as pd
import numpy as np
from . import chart


def is_evalable(obj):
    try:
        eval(str(obj))
        return True
    except:
        return False


def remove_pd_object(d):
    ret = {}
    for n, v in d.items():
        if ((not isinstance(v, pd.Series)
            and not isinstance(v, pd.DataFrame)
            and not callable(v)
            and is_evalable(v)
            ) or isinstance(v, str)):
            ret[n] = v
    return ret


def enumerate_variables(variables):
    if not variables:
        return []

    enumeration_name = []
    enumeration_vars = []

    constant_d = {}

    for name, v in variables.items():
        if (
           isinstance(v, Iterable)
           and not isinstance(v, str)
           and not isinstance(v, pd.Series)
           and not isinstance(v, pd.DataFrame)
           ):
            enumeration_name.append(name)
            enumeration_vars.append(v)
        else:
            constant_d[name] = v

    variable_enumerations = [dict(**dict(zip(enumeration_name, ps)), **constant_d)
                             for ps in list(product(*enumeration_vars))]

    return variable_enumerations


def enumerate_signal(ohlcv, strategy, variables, ):
    entries = {}
    exits = {}

    fig = {}

    iteration = tqdm.tqdm(variables) if len(variables) > 1 else variables
    for v in iteration:
        strategy.set_parameters(v)
        results = strategy.func(ohlcv)

        v = remove_pd_object(v)

        entries[str(v)], exits[str(v)] = results[0], results[1]
        if len(results) >= 3:
            fig = results[2]

    entries = pd.DataFrame(entries)
    exits = pd.DataFrame(exits)

    # setup columns
    param_names = list(eval(entries.columns[0]).keys())
    arrays = ([entries.columns.map(lambda s: eval(s)[p]) for p in param_names])
    tuples = list(zip(*arrays))
    if tuples:
        columns = pd.MultiIndex.from_tuples(tuples, names=param_names)
        exits.columns = columns
        entries.columns = columns
    return entries, exits, fig


def migrate_stop_vars(stop_vars):
    # to support OHLCSTX upgrade
    if 'ts_stop' in stop_vars:
        stop_vars['sl_stop'] = stop_vars['ts_stop']
        stop_vars['sl_trail'] = [s > 0 for s in stop_vars['ts_stop']]
    elif 'sl_stop' in stop_vars:
        stop_vars['sl_trail'] = [s < 0 for s in stop_vars['sl_stop']]
    return stop_vars


def stop_early(ohlcv, entries, exits, stop_vars, enumeration=True):
    if not stop_vars:
        return entries, exits

    # check for stop_vars
    length = -1
    stop_vars_set = {'sl_stop', 'ts_stop', 'tp_stop'}
    for s, slist in stop_vars.items():
        if s not in stop_vars_set:
            raise Exception(f'variable { s } is not one of the stop variables'
                            ': sl_stop, ts_stop, or tp_stop')
        if not isinstance(slist, Iterable):
            stop_vars[s] = [slist]

        if length == -1:
            length = len(stop_vars[s])

        if not enumeration and length != -1 and length != len(stop_vars[s]):
            raise Exception('lengths of the variables are not align: ',
                            str([len(stop_vars[s]) for s, slist in stop_vars.items()]))

    if enumeration:
        stop_vars = enumerate_variables(stop_vars)
        stop_vars = {key: [stop_vars[i][key] for i in range(len(stop_vars))] for key in stop_vars[0].keys()}

    # import pdb; pdb.set_trace()

    stop_vars = migrate_stop_vars(stop_vars)

    ohlcstx = vbt.OHLCSTX.run(
        entries,
        ohlcv['open'],
        ohlcv['high'],
        ohlcv['low'],
        ohlcv['close'],
        **stop_vars,
    )
    stop_exits = ohlcstx.exits

    nrepeat = int(len(stop_exits.columns) / len(entries.columns))
    if isinstance(stop_exits, pd.DataFrame):
        exits = exits.vbt.tile(nrepeat)
        entries = entries.vbt.tile(nrepeat)

    stop_exits = stop_exits.vbt | exits.values
    entries.columns = stop_exits.columns

    return entries, stop_exits


def plot_strategy(ohlcv, entries, exits, portfolio, fig_data, html=None):

    # format trade data
    txn = portfolio.positions.records
    txn['enter_time'] = ohlcv.iloc[txn.entry_idx].index.values
    txn['exit_time'] = ohlcv.iloc[txn.exit_idx].index.values

    # plot trade data
    mark_lines = []
    for name, t in txn.iterrows():
        x = [str(t.enter_time), str(t.exit_time)]
        y = [t.entry_price, t.exit_price]
        name = t.loc[['entry_price', 'exit_price', 'return']].to_string()
        mark_lines.append((name, x, y))

    # calculate overlap figures
    overlaps = {}
    if 'overlaps' in fig_data:
        overlaps = fig_data['overlaps']

    # calculate sub-figures
    figures = {}
    if 'figures' in fig_data:
        figures = fig_data['figures']

    figures['entries & exits'] = pd.DataFrame(
        {'entries': entries.squeeze(), 'exits': exits.squeeze()})
    figures['performance'] = portfolio.cumulative_returns()

    c, info = chart.chart(ohlcv, overlaps=overlaps,
                          figures=figures, markerlines=mark_lines,
                          start_date=ohlcv.index[-len(ohlcv)], end_date=ohlcv.index[-1])
    c.load_javascript()
    if html is not None:
        c.render(html)
    else:
        c.render()
        display(HTML(filename='render.html'))

    return


def plot_combination(portfolio, cscv_result=None, metric='final_value'):

    sns.set()
    sns.set_style("whitegrid")

    fig, axes = plt.subplots(1, 2, figsize=(15, 4), sharey=False, sharex=False)
    fig.suptitle('Backtest Results')

    def heat_map(item, name1, name2, ax):
        if name1 != name2:
            sns.heatmap(item.reset_index().pivot(name1, name2)[0], cmap='magma_r', ax=ax)
        else:
            getattr(portfolio, item).groupby(name1).mean().plot(ax=ax)

    def best_n(portfolio, n):
        return getattr(portfolio, metric)().sort_values().tail(n).index

    best_10 = best_n(portfolio, 10)

    ax = (portfolio.cumulative_returns()[best_10] * 100).plot(ax=axes[0])
    ax.set(xlabel='time', ylabel='cumulative return (%)')

    axes[1].title.set_text('Drawdown (%)')
    for n, c in zip([5, 10, 20, 30], sns.color_palette("GnBu_d")):
        bests = best_n(portfolio, n)
        drawdown = portfolio.drawdown()[bests].min(axis=1)
        ax = drawdown.plot(linewidth=1, ax=axes[1])
        # ax.fill_between(drawdown.index, 0, drawdown * 100, alpha=0.2, color=c)
    ax.set(xlabel='time', ylabel='drawdown (%)')

    plt.show()

    items = ['final_value', 'sharpe_ratio', 'sortino_ratio']
    fig, axes = plt.subplots(1, len(items), figsize=(15, 3),
                             sharey=False, sharex=False, constrained_layout=False)
    fig.subplots_adjust(top=0.75)
    fig.suptitle('Partial Differentiation')

    final_value = portfolio.final_value()
    if isinstance(final_value.index, pd.MultiIndex):
        index_names = final_value.index.names
    else:
        index_names = [final_value.index.name]

    for i, item in enumerate(items):
        results = {}
        for name in index_names:
            s = getattr(portfolio, item)()
            s = s.replace([np.inf, -np.inf], np.nan)
            results[name] = s.groupby(name).mean()
        results = pd.DataFrame(results)
        axes[i].title.set_text(item)
        results.plot(ax=axes[i])

    if cscv_result is None:
        return

    results = cscv_result

    fig, axes = plt.subplots(1, 3, figsize=(15, 5),
                             sharey=False, sharex=False, constrained_layout=False)
    fig.subplots_adjust(bottom=0.5)
    fig.suptitle('Combinatorially Symmetric Cross-validation')

    pbo_test = round(results['pbo_test'] * 100, 2)
    axes[0].title.set_text(f'Probability of overfitting: {pbo_test} %')
    axes[0].hist(x=[l for l in results['logits'] if l > -10000], bins='auto')
    axes[0].set_xlabel('Logits')
    axes[0].set_ylabel('Frequency')

    # performance degradation
    axes[1].title.set_text('Performance degradation')
    x, y = pd.DataFrame([results['R_n_star'], results['R_bar_n_star']]).dropna(axis=1).values
    sns.regplot(x, y, ax=axes[1])
    # axes[1].set_xlim(min(results['R_n_star']) * 1.2,max(results['R_n_star']) * 1.2)
    # axes[1].set_ylim(min(results['R_bar_n_star']) * 1.2,max(results['R_bar_n_star']) * 1.2)
    axes[1].set_xlabel('In-sample Performance')
    axes[1].set_ylabel('Out-of-sample Performance')

    # first and second Stochastic dominance
    axes[2].title.set_text('Stochastic dominance')
    if len(results['dom_df']) != 0:
        results['dom_df'].plot(ax=axes[2], secondary_y=['SD2'])
    axes[2].set_xlabel('Performance optimized vs non-optimized')
    axes[2].set_ylabel('Frequency')


def variable_visualization(portfolio):

    param_names = portfolio.cumulative_returns().columns.names
    dropdown1 = widgets.Dropdown(
        options=param_names,
        value=param_names[0],
        description='axis 1:',
        disabled=False,
    )
    dropdown2 = widgets.Dropdown(
        options=param_names,
        value=param_names[0],
        description='axis 2:',
        disabled=False,
    )

    performance_metric = [
        'final_value',
        'calmar_ratio', 'max_drawdown', 'sharpe_ratio',
        'downside_risk', 'omega_ratio', 'conditional_value_at_risk']

    performance_dropdwon = widgets.Dropdown(
        options=performance_metric,
        value=performance_metric[0],
        description='performance',
        disabled=False,
    )

    out = widgets.Output()

    def update(v):
        name1 = dropdown1.value
        name2 = dropdown2.value
        performance = performance_dropdwon.value

        with out:
            out.clear_output()
            if name1 != name2:
                df = (getattr(portfolio, performance)()
                      .reset_index().groupby([name1, name2]).mean()[0]
                      .reset_index().pivot(name1, name2)[0])

                df = df.replace([np.inf, -np.inf], np.nan)
                sns.heatmap(df)
            else:
                getattr(portfolio, performance)().groupby(name1).mean().plot()
            plt.show()

    dropdown1.observe(update, 'value')
    dropdown2.observe(update, 'value')
    performance_dropdwon.observe(update, 'value')
    drawdowns = widgets.VBox([
        performance_dropdwon,
        widgets.HBox([dropdown1, dropdown2])])
    display(drawdowns)
    display(out)
    update(0)



from crypto_strategy.strategies.bo_strategy import InspectBoStrategy
from crypto_strategy.strategies.ma_strategy import InspectMaStrategy


def calc_kelly_criterion(stats):
    # winning probability
    p = stats['Win Rate [%]'] / 100
    # the odds defining the amount won for unit bet
    b = abs(stats['Avg Winning Trade [%]'] / stats['Avg Losing Trade [%]'])
    # optimal share of capital to bet (b * p + p - 1) / b
    kc = p + (p - 1) / b
    return kc


def tm_bo_stats(symbol, tm):
    parts = tm.split('-')
    freq = parts[1]
    strategy = parts[2]
    long_window = int(parts[3])
    short_window = int(parts[4])
    flag_filter = timeperiod = multiplier = threshold = stop_vars = None
    if 'vol' in parts:
        flag_filter = parts[-4]
        timeperiod = int(parts[-3])
        multiplier = int(parts[-2])
    if 'ang' in parts:
        flag_filter = parts[-4]
        timeperiod = int(parts[-3])
        threshold = int(parts[-2])
    if 'ts' in parts:
        stop_vars = {'ts_stop': float(parts[6])}
    if 'sl' in parts:
        stop_vars = {'sl_stop': float(parts[6])}
    if 'tp' in parts:
        stop_vars = {'tp_stop': float(parts[6])}
    ins = InspectBoStrategy(
        symbol,
        freq=freq,
        long_window=long_window,
        short_window=short_window,
        timeperiod=timeperiod,
        multiplier=multiplier,
        threshold=threshold,
        flag_filter=flag_filter,
        stop_vars=stop_vars,
        strategy=strategy,
        show_fig=False
    )
    return ins.portfolio


def tm_ma_stats(symbol, tm):
    parts = tm.split('-')
    freq = parts[1]
    name = parts[2]
    n1 = int(parts[3])
    n2 = int(parts[4])
    flag_filter = timeperiod = stop_vars = None
    if 'mmi' in parts or 'mmi_ge' in parts:
        flag_filter = parts[-3]
        timeperiod = int(parts[-2])
    if 'ts' in parts:
        stop_vars = {'ts_stop': float(parts[6])}
    if 'sl' in parts:
        stop_vars = {'sl_stop': float(parts[6])}
    if 'tp' in parts:
        stop_vars = {'tp_stop': float(parts[6])}
    ins = InspectMaStrategy(
        symbol,
        freq=freq,
        name=name,
        n1=n1,
        n2=n2,
        flag_filter=flag_filter,
        timeperiod=timeperiod,
        stop_vars=stop_vars,
        show_fig=False
    )
    return ins.portfolio


def portfolio_stats(tm_list):
    tp_stats = dict()
    for tm in tm_list:
        key = tm.name
        symbol = tm.symbols[0]
        tp_stats[key] = dict()
        stats = None
        if 'bo' in tm.name:
            stats = tm_bo_stats(symbol, tm.name)
        else:
            stats = tm_ma_stats(symbol, tm.name)
        if stats is not None:
            stats = stats.stats()
            tp_stats[key] = stats
    return tp_stats


def portfolio_kelly_criterion(tm_list):
    tp_kc = dict()
    for tm in tm_list:
        key = tm.name
        symbol = tm.symbols[0]
        tp_kc[key] = dict()
        stats = None
        if 'bo' in tm.name:
            stats = tm_bo_stats(symbol, tm.name)
        else:
            stats = tm_ma_stats(symbol, tm.name)
        if stats is not None:
            stats = stats.stats()
            tp_kc[key]['trades'] = stats['Total Trades']
            tm_kc = calc_kelly_criterion(stats)
            tp_kc[key]['kc'] = tm_kc
        else:
            raise ValueError(f'stats not found for {key}')
    return tp_kc


def multi_assets_allocation(tm_list):
    tp_kc = dict()
    for tm in tm_list:
        key = tm.name
        symbol = tm.symbols[0]
        tp_kc[key] = dict()
        stats = None
        if 'bo' in tm.name:
            stats = tm_bo_stats(symbol, key)
        else:
            stats = tm_ma_stats(symbol, key)
        if stats is not None:
            stats = stats.annual_returns()
            stats.index = stats.index.year
            tp_kc[key] = stats
        else:
            raise ValueError(f'stats not found for {key}')
    return tp_kc


def portfolio_monthly_ret(tm_list):
    tp_monthly_ret = dict()
    for tm in tm_list:
        key = tm.name
        symbol = tm.symbols[0]
        tp_monthly_ret[key] = dict()
        stats = None
        if 'bo' in tm.name:
            stats = tm_bo_stats(symbol, tm.name)
        else:
            stats = tm_ma_stats(symbol, tm.name)
        if stats is not None:
            tp_monthly_ret[key] = stats.get_qs().monthly_returns()
    return tp_monthly_ret


def portfolio_daily_ret(tm_list):
    tp_daily_ret = dict()
    for tm in tm_list:
        key = tm.name
        symbol = tm.symbols[0]
        tp_daily_ret[key] = dict()
        stats = None
        if 'bo' in tm.name:
            stats = tm_bo_stats(symbol, tm.name)
        else:
            stats = tm_ma_stats(symbol, tm.name)
        if stats is not None:
            tp_daily_ret[key] = stats.daily_returns()
    return tp_daily_ret



'''
The code from finlab folder is originally from https://github.com/finlab-python/finlab_crypto
'''
import vectorbt as vbt


# set default fees and slippage
vbt.settings.portfolio['init_cash'] = 100.0  # in $
vbt.settings.portfolio['fees'] = 0.001       # in %
vbt.settings.portfolio['slippage'] = 0.001   # in %



import os
from abc import ABC, abstractmethod
import pandas as pd
from finlab_crypto.overfitting import CSCV


def generate_cscv(portfolio, cscv_nbins=10, cscv_objective=lambda r: r.mean()):
    cscv = CSCV(n_bins=cscv_nbins, objective=cscv_objective)
    cscv.add_daily_returns(portfolio.daily_returns())
    cscv_result = cscv.estimate_overfitting(plot=False)
    return cscv_result


class GenerateReport(ABC):
    def __init__(self, symbols: list, date: str, res_dir: str):
        self.symbols = symbols
        self.date = date
        self.res_dir = os.path.join(res_dir, date.strip('-'))
        self.generate_report()

    @abstractmethod
    def print_params(self, params):
        pass

    def generate_report(self):
        if os.path.isdir(self.res_dir) and self.symbols:
            files = os.listdir(self.res_dir)
            print(f'{self.date} update')
            for symbol in self.symbols:
                for f in files:
                    if f.startswith(symbol) and f.endswith(self.date + '.pkl'):
                        res = pd.read_pickle(os.path.join(self.res_dir, f))
                        params = f.split('-')
                        self.print_params(params)
                        res = res.reset_index()
                        res.columns = ['', '']
                        print(res.to_markdown(showindex=False), '\n')
                        # display(HTML(rslt.reset_index().to_html(header=False)))



import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from crypto_strategy.data import join_path
from .base import generate_cscv


def plot_indicators(portfolio, output_path=None, filename=None, show_fig=False):
    items = ['final_value', 'sharpe_ratio', 'sortino_ratio']
    fig, axes = plt.subplots(1, len(items), figsize=(15, 3),
                             sharey=False, sharex=False, constrained_layout=False)
    fig.subplots_adjust(top=0.75)
    fig.suptitle('Partial Differentiation')

    final_value = portfolio.final_value()
    if isinstance(final_value.index, pd.MultiIndex):
        index_names = final_value.index.names
    else:
        index_names = [final_value.index.name]

    for i, item in enumerate(items):
        results = {}
        for name in index_names:
            s = getattr(portfolio, item)()
            s = s.replace([np.inf, -np.inf], np.nan)
            results[name] = s.groupby(name).mean()
        results = pd.DataFrame(results)
        axes[i].title.set_text(item)
        results.plot(ax=axes[i])

    if output_path and filename:
        file_diff = join_path(output_path, filename + '-diff.png')
        plt.savefig(file_diff)

    results = generate_cscv(portfolio)
    fig, axes = plt.subplots(1, 3, figsize=(15, 5),
                             sharey=False, sharex=False, constrained_layout=False)
    fig.subplots_adjust(bottom=0.5)
    fig.suptitle('Combinatorially Symmetric Cross-validation')

    pbo_test = round(results['pbo_test'] * 100, 2)
    axes[0].title.set_text(f'Probability of overfitting: {pbo_test} %')
    axes[0].hist(x=[r for r in results['logits'] if r > -10000], bins='auto')
    axes[0].set_xlabel('Logits')
    axes[0].set_ylabel('Frequency')

    # performance degradation
    axes[1].title.set_text('Performance degradation')
    x, y = pd.DataFrame([results['R_n_star'], results['R_bar_n_star']]).dropna(axis=1).values
    sns.regplot(x, y, ax=axes[1])
    # axes[1].set_xlim(min(results['R_n_star']) * 1.2,max(results['R_n_star']) * 1.2)
    # axes[1].set_ylim(min(results['R_bar_n_star']) * 1.2,max(results['R_bar_n_star']) * 1.2)
    axes[1].set_xlabel('In-sample Performance')
    axes[1].set_ylabel('Out-of-sample Performance')

    # first and second Stochastic dominance
    axes[2].title.set_text('Stochastic dominance')
    if len(results['dom_df']) != 0:
        results['dom_df'].plot(ax=axes[2], secondary_y=['SD2'])
    axes[2].set_xlabel('Performance optimized vs non-optimized')
    axes[2].set_ylabel('Frequency')

    if output_path and filename:
        file_cscv = join_path(output_path, filename + '-cscv.png')
        plt.savefig(file_cscv)

    if show_fig:
        plt.show()


def plot_returns(btc_acc_returns, eth_acc_returns, pf_acc_returns, start_date):
    fig, ax = plt.subplots()
    btc_acc_returns.plot(figsize=(8, 6), label='BTC')
    ax.annotate(f'{btc_acc_returns[-1]:.2f}', (btc_acc_returns.index[-1], btc_acc_returns[-1]))
    eth_acc_returns.plot(label='ETH')
    ax.annotate(f'{eth_acc_returns[-1]:.2f}', (eth_acc_returns.index[-1], eth_acc_returns[-1]))
    pf_acc_returns.plot(label='Portfolio')
    ax.annotate(f'{pf_acc_returns[-1]:.2f}', (pf_acc_returns.index[-1], pf_acc_returns[-1]))
    ax.legend()
    ax.set_title(f'{start_date}-day Portfolio Return')
    fig.tight_layout()



from typing import Union
import numpy as np
import pandas as pd
from crypto_strategy.data import (
    save_stats,
    get_acc_returns,
    download_crypto_history
)
from .base import (
    breakout_strategy,
    vol_filter, ang_filter,
    BestStrategy, InspectStrategy, CheckIndicators
)


RANGE_WINDOW = np.arange(10, 200, 5)
RANGE_TIMEPERIOD = np.arange(5, 55, 5)
RANGE_MULTIPLIER = np.arange(1, 10, 1)
RANGE_THRESHOLD = np.arange(0, 60, 5)
RANGE_STOP = np.arange(0.1, 0.6, 0.05)


def create_vol_filter(flag_filter, **kwargs):
    filters = dict()
    if flag_filter:
        if kwargs:
            assert kwargs.get('timeperiod') and kwargs.get('multiplier'),\
                'Volume fitler doesn\'t have config params provided'
            filter_config = dict(
                timeperiod=kwargs['timeperiod'],
                multiplier=kwargs['multiplier']
            )
        else:
            filter_config = dict(
                timeperiod=RANGE_TIMEPERIOD,
                multiplier=RANGE_MULTIPLIER
            )
        filters = {
            'vol': vol_filter.create({
                'timeperiod': filter_config['timeperiod'],
                'multiplier': filter_config['multiplier']
            })
        }
    return filters


def create_ang_filter(flag_filter, **kwargs):
    filters = dict()
    if flag_filter:
        if kwargs:
            assert kwargs.get('timeperiod') and kwargs.get('threshold') is not None,\
                'Angle filter doesn\'t have config params provided'
            filter_config = dict(
                timeperiod=kwargs['timeperiod'],
                threshold=kwargs['threshold']
            )
        else:
            filter_config = dict(
                timeperiod=RANGE_TIMEPERIOD,
                threshold=RANGE_THRESHOLD
            )
        filters = {
            'ang': ang_filter.create({
                'timeperiod': filter_config['timeperiod'],
                'threshold': filter_config['threshold']
            })
        }
    return filters


def get_filter(flag_filter, **kwargs):
    filters = dict()
    if flag_filter == 'vol':
        filters = create_vol_filter(flag_filter, **kwargs)
    if flag_filter == 'ang':
        filters = create_ang_filter(flag_filter, **kwargs)
    return filters


def get_strategy(strategy):
    if strategy == 'bo':
        return breakout_strategy
    else:
        raise ValueError('Strategy not recognized. Please choose from bo and bo_rev')


def create_bo_variables(**kwargs):
    variables = dict()
    if kwargs.get('long_window') and kwargs.get('short_window'):
        variables = kwargs
    elif not (kwargs.get('long_window') or kwargs.get('short_window')):
        variables['long_window'] = RANGE_WINDOW
        variables['short_window'] = RANGE_WINDOW
    else:
        raise ValueError('BO strategy doesn\'t have config params provided')
    return variables


def create_bo_variables_with_stop(**kwargs):
    variables = create_bo_variables(**kwargs)
    if not (kwargs.get('long_window') or kwargs.get('short_window')):
        flag_stop = kwargs.get('flag_stop')
        if kwargs.get('flag_stop'):
            if not set(flag_stop).issubset(['ts_stop', 'sl_stop', 'tp_stop']):
                raise ValueError('The value of flag_stop is not supported')
            for fs in flag_stop:
                variables[fs] = RANGE_STOP
        else:
            raise ValueError('flag_stop is not set')
    return variables


class BestBoStrategy(BestStrategy):
    '''
    This class provides the method to optimize the BO strategy
    symbols: a list of symbols to be optimzied on, e.g., ['BTCUSDT']
    freq: currently supported values: '1h' or '4h'
    res_dir: the output directory
    flag_filter: currently supported fitlers: 'vol', 'ang', default: None
    flag_ts_stop: flag to turn on/off trailing stop
    strategy: 'bo'
    '''
    def __init__(self,
                 symbols: Union[str, list],
                 freq: str,
                 res_dir: str,
                 flag_filter: str = None,
                 flag_stop: Union[str, list] = None,
                 flag_acc_return: bool = True,
                 strategy: str = 'bo',
                 ):
        super().__init__(symbols, freq, res_dir, flag_filter, strategy)
        self.flag_stop = [flag_stop] if isinstance(flag_stop, str) else flag_stop
        self.flag_acc_return = flag_acc_return
        self.generate_best_params()

    def _get_strategy(self, strategy):
        return get_strategy(strategy)

    def _get_filter(self, **kwargs):
        return get_filter(flag_filter=self.flag_filter, **kwargs)

    def _get_variables(self, **kwargs):
        if self.flag_stop:
            return create_bo_variables_with_stop(flag_stop=self.flag_stop, **kwargs)
        return create_bo_variables(**kwargs)

    def _get_grid_search(self):
        if self.flag_filter == 'vol':
            return self.grid_search_vol_params()
        elif self.flag_filter == 'ang':
            return self.grid_search_ang_params()
        else:
            return self.grid_search_params()

    def get_best_params(self, total_best_params, n=10):
        total_best_params = pd.DataFrame(total_best_params)
        total_best_params['sharpe'] = (total_best_params['sharpe'] + 0.05).round(1)
        total_best_params['gap'] = total_best_params['long_window'] - total_best_params['short_window']
        if 'ohlcstx_sl_stop' in total_best_params.columns:
            total_best_params = (
                total_best_params
                .query('gap > 0')
                .sort_values(
                    by=['sharpe', 'gap', 'short_window', 'ohlcstx_sl_stop'],
                    ascending=[True, True, False, False])
            )
        else:
            total_best_params = (
                total_best_params
                .query('gap > 0')
                .sort_values(
                    by=['sharpe', 'gap', 'short_window'],
                    ascending=[True, True, False]
                    )
            )
        print(total_best_params)
        return total_best_params.tail(1).to_dict(orient='records')[0] if not total_best_params.empty else None

    def grid_search_params(self):
        variables = self._get_variables()
        best_params = self.backtest(variables)
        return self.get_best_params(best_params)

    def grid_search_vol_params(self):
        variables = self._get_variables()
        total_best_params = list()
        for multiplier in RANGE_MULTIPLIER:
            for timeperiod in RANGE_TIMEPERIOD:
                filters = self._get_filter(timeperiod=timeperiod, multiplier=multiplier)
                best_params = self.backtest(variables, filters)
                total_best_params.extend(best_params)
        return self.get_best_params(total_best_params)

    def grid_search_ang_params(self):
        variables = self._get_variables()
        total_best_params = list()
        for threshold in RANGE_THRESHOLD:
            for timeperiod in RANGE_TIMEPERIOD:
                filters = self._get_filter(timeperiod=timeperiod, threshold=threshold)
                best_params = self.backtest(variables, filters)
                total_best_params.extend(best_params)
        return self.get_best_params(total_best_params)

    def apply_best_params(self, best_params, symbol):
        variables = self._get_variables(
            long_window=best_params['long_window'],
            short_window=best_params['short_window']
        )
        if 'ts_stop' in self.flag_stop:
            variables.update(ts_stop=best_params.get('ohlcstx_sl_stop'))
        if 'sl_stop' in self.flag_stop:
            variables.update(sl_stop=best_params.get('ohlcstx_sl_stop'))
        if 'tp_stop' in self.flag_stop:
            variables.update(tp_stop=best_params.get('ohlcstx_tp_stop'))
        filters = self._get_filter(
            timeperiod=best_params.get('vol_timeperiod') if best_params.get('vol_timeperiod') else best_params.get('ang_timeperiod'),
            multiplier=best_params.get('vol_multiplier'),
            threshold=best_params.get('ang_threshold')
            )
        filename = f'''{symbol}-{self.freq}-{self.strategy_name}-{best_params['long_window']}-{best_params['short_window']}-'''
        if self.flag_stop:
            if 'sl_stop' in self.flag_stop:
                filename += f"sl_stop-{best_params.get('ohlcstx_sl_stop'):.2f}-"
            if 'ts_stop' in self.flag_stop:
                filename += f"ts_stop-{best_params.get('ohlcstx_sl_stop'):.2f}-"
            if 'tp_stop' in self.flag_stop:
                filename += f"tp_stop-{best_params.get('ohlcstx_tp_stop'):.2f}-"
        if self.flag_filter == 'vol':
            filename += f'''{self.flag_filter}-{best_params['vol_timeperiod']}-{best_params['vol_multiplier']}-{self.date_str}.pkl'''
        elif self.flag_filter == 'ang':
            filename += f'''{self.flag_filter}-{best_params['ang_timeperiod']}-{best_params['ang_threshold']}-{self.date_str}.pkl'''
        else:
            filename += f'''{self.date_str}.pkl'''
        portfolio = self.strategy.backtest(self.ohlcv, freq=self.freq, variables=variables, filters=filters)
        stats = portfolio.stats()
        if self.flag_acc_return:
            acc_returns = get_acc_returns(portfolio.daily_returns())
            stats = stats.append(pd.Series(acc_returns))
        save_stats(stats, self.output_path, filename)
        print(f'The stats are saved to {self.output_path}/{filename}')

    def generate_best_params(self):
        print(f'The search for {self.symbols} starts now')
        for symbol in self.symbols:
            print(f'Start the search for {symbol}')
            self.ohlcv = download_crypto_history(symbol, self.freq)
            best_params = self._get_grid_search()
            if best_params:
                print(f'Found the best params {best_params}')
                self.apply_best_params(best_params, symbol)
            else:
                print('Not found the best params')
        print('The search for all the symbols is completed')


class CheckBoIndicators(CheckIndicators):
    '''
    This class provides the method to check Partial Differentiation
    and Combinatorially Symmetric Cross-validation of BO strategy.
    symbols: a list of symbols to be optimzied on, e.g., ['BTCUSDT']
    date: the date the best params are created
    res_dir: the output directory
    flag_filter: currently supported fitlers: 'vol', 'ang', default: None
    flag_ts_stop: flag to turn on/off trailing stop
    strategy: currently supported values: 'bo'
    '''
    def __init__(self,
                 symbols: Union[str, list],
                 date: str,
                 res_dir: str,
                 flag_filter: str = None,
                 flag_stop: Union[str, list] = None,
                 strategy: str = 'bo'
                 ):
        super().__init__(symbols, date, res_dir, flag_filter, strategy)
        self.flag_stop = [flag_stop] if isinstance(flag_stop, str) else flag_stop
        self.check_indicators()

    def _get_strategy(self, strategy):
        return get_strategy(strategy)

    def _get_variables(self, **kwargs):
        if self.flag_stop:
            return create_bo_variables_with_stop(flag_stop=self.flag_stop, **kwargs)
        return create_bo_variables(**kwargs)

    def _get_filter(self, **kwargs):
        return get_filter(flag_filter=self.flag_filter, **kwargs)


class InspectBoStrategy(InspectStrategy):
    '''
    This class provides the method to optimize the BO strategy
    symbols: a list of symbols to be optimzied on, e.g., 'BTCUSDT'
    freq: currently supported values: '1h' or '4h'
    long_window, short_window: breakout params
    timeperiod, multiplier: volume filter params
    flag_filter: currently supported fitlers: 'vol', 'ang', default: None
    ts_stop: ts_stop params
    strategy: currently supports 'bo'
    '''
    def __init__(self, 
                 symbol: str,
                 freq: str,
                 long_window: int,
                 short_window: int,
                 timeperiod: int = None,
                 multiplier: int = None,
                 threshold: int = None,
                 flag_filter: str = None,
                 stop_vars: dict = None,
                 strategy: str = 'bo',
                 show_fig: bool = True
                 ):
        super().__init__(symbol, freq, flag_filter, strategy, show_fig)
        self.long_window = long_window
        self.short_window = short_window
        self.timeperiod = timeperiod
        self.multiplier = multiplier
        self.threshold = threshold
        self.stop_vars = stop_vars
        self.inspect()

    def _get_strategy(self, strategy):
        return get_strategy(strategy)

    def _get_variables(self, **kwargs):
        variables = create_bo_variables(
            long_window=self.long_window,
            short_window=self.short_window
        )
        if self.stop_vars:
            variables.update(self.stop_vars)
        return variables

    def _get_filter(self):
        return get_filter(
            self.flag_filter,
            timeperiod=self.timeperiod,
            multiplier=self.multiplier,
            threshold=self.threshold
            )


def returns_timeline(
    symbol: str,
    freq: str,
    long_window: int,
    short_window: int,
    timeperiod: int = None,
    multiplier: int = None,
    threshold: int = None,
    flag_filter: str = None,
    stop_vars: dict = None,
    strategy: str = 'bo',
):
    ins = InspectBoStrategy(
        symbol=symbol,
        freq=freq,
        long_window=long_window,
        short_window=short_window,
        timeperiod=timeperiod,
        multiplier=multiplier,
        threshold=threshold,
        flag_filter=flag_filter,
        stop_vars=stop_vars,
        strategy=strategy,
        show_fig=False
    )
    daily_returns = ins.portfolio.daily_returns()
    acc_returns = get_acc_returns(daily_returns)
    return acc_returns



import numpy as np
import pandas as pd
import pyecharts.options as opts
from pyecharts.globals import CurrentConfig, NotebookType
from pyecharts.charts import Kline, Line, Grid, Bar
CurrentConfig.NOTEBOOK_TYPE = NotebookType.JUPYTER_LAB


def chart(dfstock, overlaps=dict(), figures=dict(), markers=dict(), markerlines=[], start_date=None, end_date=None):
    """Backtesting Analysis and optimizer dashboard platform.

    Use pyechart and seaborn module to generate interactive variety charts.

    Args:
      dfstock: A dataframe of trading target data.
      overlaps: A dict of overlaps indicator line setting in figure.
      figures: A dict of information needed for picture drawing.
      markers: A dict of which dfstock index needed to be mark.
      markerlines: A tuple(name, x, y ) in dict of drawing the line connection between entry to exist point.
      start_date: A datetime value of the start of dfstock.
      end_date: A datetime value of the end of dfstock .

    Returns:
      grid_chart: chart display.
      chart_size: A dict of chart's height and width values.

    """
    title = 60
    title_margin_top = 30
    main_chart_height = 300
    margin_left = 50
    vol_chart_height = 50
    sub_figure_height = 60
    width = 800

    dfstock = dfstock.loc[start_date:end_date]

    mark_data = []
    for mark in markers:

        if mark[1] not in dfstock.index:
            continue

        x = np.where(dfstock.index == mark[1])[0][0]
        y = dfstock.high.loc[mark[1]]
        color = '#1d6ff2'
        o = opts.MarkPointItem(coord=[float(x), y], value=mark[0], itemstyle_opts=opts.ItemStyleOpts(color=color))
        mark_data.append(o)

    modified_marklines = []
    for markline in markerlines:
        name, x, y = markline
        if x[0] not in dfstock.index or x[1] not in dfstock.index:
            continue
        xx0 = np.where(dfstock.index == x[0])[0][0]
        xx1 = np.where(dfstock.index == x[1])[0][0]
        x = [float(xx0), float(xx1)]
        modified_marklines.append([
            {
                'name': name,
                'coord': [x[0], y[0]],
                'itemStyle': {'color': '#216dc4'}
            },
            {
                'coord': [x[1], y[1]]
            }
            ])

    # for m in modified_marklines:
    #  print(m.opts)
    #  print('------')

    # mark_data += [
    #     opts.MarkPointItem(type_="max", name="最大值", symbol='rect', symbol_size=[50, 20],
    #                        itemstyle_opts=opts.ItemStyleOpts(color='rgba(0,0,0,0.3)')
    #                       ),
    #     opts.MarkPointItem(type_="min", name="最小值", symbol='rect', symbol_size=[50, 20],
    #                        itemstyle_opts=opts.ItemStyleOpts(color='rgba(0,0,0,0.3)')
    #     )
    # ]

    kline = (
        Kline()
        .add_xaxis(xaxis_data=dfstock.index.astype(str).to_list())
        .add_yaxis(
            series_name="klines",
            y_axis=dfstock[['open', 'close', 'low', 'high']].values.tolist(),
            markpoint_opts=opts.MarkPointOpts(
                data=mark_data
            ),
            markline_opts=opts.MarkLineOpts(
                data=modified_marklines,
                label_opts={'position': 'insideMiddleTop', 'show': False}
            ),
            itemstyle_opts=opts.ItemStyleOpts(
                color="#ff6183",
                color0="#58d6ac",
                border_color="#ff6183",
                border_color0="#58d6ac",
            ),
        )
        .set_series_opts()
    )

    #################
    # overlap chart
    #################

    overlap_chart = (
        Line()
        .add_xaxis(xaxis_data=dfstock.index.astype(str).to_list())
    )
    for name, o in overlaps.items():
        overlap_chart.add_yaxis(
            series_name=name,
            y_axis=o.loc[start_date:end_date].to_list(),
            is_smooth=True,
            is_hover_animation=False,
            linestyle_opts=opts.LineStyleOpts(opacity=0.5),
            label_opts=opts.LabelOpts(is_show=False),
        )

    # Bar-1
    bar_1 = (
        Bar()
        .add_xaxis(xaxis_data=dfstock.index.astype(str).to_list())
        .add_yaxis(
            series_name="volume",
            y_axis=dfstock.volume.loc[start_date:end_date].to_list(),
            xaxis_index=1,
            yaxis_index=1,
            label_opts=opts.LabelOpts(is_show=False),
            # 改进后在 grid 中 add_js_funcs 后变成如下
            itemstyle_opts=opts.ItemStyleOpts(
                color='rgba(0,0,0,0.2)',
            ),
        )
        .set_global_opts(
            xaxis_opts=opts.AxisOpts(
                type_="category",
                grid_index=1,
                axislabel_opts=opts.LabelOpts(is_show=False),
            ),
            yaxis_opts=opts.AxisOpts(
                axislabel_opts=opts.LabelOpts(is_show=False),
            ),
            legend_opts=opts.LegendOpts(is_show=False),
        )
    )

    def is_item(item):
        return isinstance(item, pd.Series) or isinstance(item, tuple)

    def item_to_chart(name, item):
        if isinstance(item, pd.Series):
            item_type = 'line'
            series = item.loc[start_date:end_date]
        elif isinstance(item, tuple):
            item_type = item[1]
            series = item[0].loc[start_date:end_date]
        else:
            print('Object type not accept (only pd.Series or tuple)')
            raise

        values = series.to_list()
        index = series.index.astype(str).to_list()

        chart = None
        if item_type == 'line':
            chart = Line()
            chart.add_xaxis(xaxis_data=index)
            chart.add_yaxis(
                series_name=name,
                y_axis=values,
                is_hover_animation=False,
                # linestyle_opts=opts.LineStyleOpts(width=3, opacity=0.5),
                label_opts=opts.LabelOpts(is_show=False),
                )
        elif item_type == 'bar':
            chart = Bar()
            chart.add_xaxis(xaxis_data=index)
            chart.add_yaxis(
                series_name=name,
                y_axis=values,
                # xaxis_index=1,
                # yaxis_index=1,
                label_opts=opts.LabelOpts(is_show=False),
                )

        return chart

    example_charts = []
    for name, graph in figures.items():
        if is_item(graph):
            example_charts.append(item_to_chart(name, graph))
        elif isinstance(graph, dict) or isinstance(graph, pd.DataFrame):
            ys = [item_to_chart(name, subgraph) for name, subgraph in graph.items()]
            for y in ys[1:]:
                ys[0].overlap(y)
            example_charts.append(ys[0])
        else:
            raise Exception('cannot support subfigure type')

    if len(dfstock) <= 500:
        range_start = 0
    else:
        range_start = 95  # 100 - int(10000/len(dfstock))

    kline.set_global_opts(
            legend_opts=opts.LegendOpts(pos_top='0px', pos_left=str(margin_left)),
            xaxis_opts=opts.AxisOpts(is_scale=True),
            yaxis_opts=opts.AxisOpts(
                is_scale=True,
                splitarea_opts=opts.SplitAreaOpts(
                    is_show=True, areastyle_opts=opts.AreaStyleOpts(opacity=0.3)
                ),
                # grid_index=1,
                # split_number=3,
                axisline_opts=opts.AxisLineOpts(is_on_zero=False),
                # axistick_opts=opts.AxisTickOpts(is_show=False),
                splitline_opts=opts.SplitLineOpts(is_show=False),
                axislabel_opts=opts.LabelOpts(is_show=True),
            ),
            datazoom_opts=[
                opts.DataZoomOpts(
                    is_show=False,
                    type_="inside",
                    xaxis_index=list(range(len(example_charts)+2)),
                    range_start=range_start,
                    range_end=100,
                ),
                opts.DataZoomOpts(
                    is_show=True,
                    xaxis_index=list(range(len(example_charts)+2)),
                    type_="slider",
                    pos_top="85%",
                    range_start=range_start,
                    range_end=100,
                ),
            ],
            # title_opts=opts.TitleOpts(title="Kline-DataZoom-inside"),
        )

    # Kline And Line
    overlap_kline_line = kline.overlap(overlap_chart)

    total_height = title + main_chart_height + len(example_charts) * (sub_figure_height + title) + 200

    # Grid Overlap + Bar
    grid_chart = Grid(
        init_opts=opts.InitOpts(
            width=str(width) + 'px',
            height=str(total_height) + 'px',
            animation_opts=opts.AnimationOpts(animation=False),
        )
    )
    grid_chart.add(
        overlap_kline_line,
        grid_opts=opts.GridOpts(pos_top=str(title) + 'px',
                                height=str(main_chart_height) + 'px',
                                pos_left=str(margin_left)+'px', pos_right='0'),
    )

    grid_chart.add(
        bar_1,
        grid_opts=opts.GridOpts(pos_top=str(title+main_chart_height-vol_chart_height) + 'px',
                                height=str(vol_chart_height) + 'px',
                                pos_left=str(margin_left)+'px', pos_right='0'),
    )

    for i, chart in enumerate(example_charts):
        title_pos_top = title + main_chart_height + i * (sub_figure_height + title)
        chart.set_global_opts(
                # title_opts=opts.TitleOpts(name, pos_top=str(title_pos_top+title_margin_top) + 'px'),
                legend_opts=opts.LegendOpts(pos_left=str(margin_left), pos_top=str(title_pos_top+title_margin_top) + 'px'),
            )
        chart_pos_top = title_pos_top + title
        grid_chart.add(
            chart,
            grid_opts=opts.GridOpts(pos_top=str(chart_pos_top) + 'px',
                                    height=str(sub_figure_height) + 'px',
                                    pos_left=str(margin_left)+'px', pos_right='0'
                                    ),
        )
        chart_size = {'height': total_height,  'width': width}
    return grid_chart, chart_size



"""Strategy function plug-in.

You can use Filter and Strategy function as decorator
that make strategies easy to construct filters layers
and common strategy detection methods, such as back-testing,
parameter tuning and analysis charts.

  Typical usage example:
```
  @Filter(timeperiod=20)
  def your_filter(ohlcv):
      your filter logic...
      return filter > filter_value, figures
  f60 = your_filter.create({'timeperiod': 60})
```
  -------------------------------
```
  @Strategy()
  def your_strategy(ohlcv):
     your strategy logic...
      return entries, exits, figures
  portfolio = your_strategy.backtest(ohlcv, freq='4h', plot=True)
```
"""

from finlab_crypto.utility import (enumerate_variables, enumerate_signal,
                                   stop_early, plot_combination, plot_strategy,
                                   variable_visualization, remove_pd_object
                                   )
from finlab_crypto.overfitting import CSCV
import copy
import vectorbt as vbt
import pandas as pd
import matplotlib.pyplot as plt
from collections.abc import Iterable


class Filter(object):
    """Filter package features plug-in.

    Offer easy way to create filter to use in class Strategy.

    Attributes:
        default_parameters: customized filter attributes.
    """

    def __init__(self, **default_parameters):
        """inits filter."""
        self.func = None
        self.filters = {}
        self._default_parameters = default_parameters
        self.set_parameters(default_parameters)

    def __call__(self, func):
        """decorator function

        Args
          func: A function of the customized filter.
        """
        self.func = func
        return self

    def set_parameters(self, variables):
        """set your customized filter parameters.

        let filter class use variables dict to set method

        Args:
          variables: a dict of your customized filter attributes.

        """
        if self._default_parameters:
            for key, val in self._default_parameters.items():
                setattr(self, key, val)

        if variables:
            for key, val in variables.items():
                setattr(self, key, val)

    def show_parameters(self):
        parameters = {}
        for key, val in self._default_parameters.items():
            parameters[key] = getattr(self, key)
        print(parameters)

    def create(self, variables=None):
        """generate filter signals, fig_data.

        offer easy way to create filter signals, fig_data

        Args:
          variables: a dict of your customized filter attributes.
        Returns:
          signals: a dataframe of filter signals.
          fig_data: a dict of required data for figure display.
        """

        def ret_f(ohlcv):

            variable_enumerate = enumerate_variables(variables)
            if len(variable_enumerate) == 0:
                variable_enumerate.append(self._default_parameters)

            signals = {}
            fig_data = {}
            for v in variable_enumerate:

                self.set_parameters(v)
                results = self.func(ohlcv)

                v = remove_pd_object(v)

                if isinstance(results, Iterable):
                    signals[str(v)], fig_data = results
                else:
                    signals[str(v)] = results

            signals = pd.DataFrame(signals)
            signals.columns.name = 'filter'

            param_names = list(eval(signals.columns[0]).keys())
            arrays = ([signals.columns.map(lambda s: eval(s)[p]) for p in param_names])
            tuples = list(zip(*arrays))
            columns = pd.MultiIndex.from_tuples(tuples, names=param_names)
            signals.columns = columns

            return signals, fig_data

        return ret_f


class Strategy(object):
    """strategy features plug-in.

    offer common strategy detection methods, such as back-testing,
    parameter tuning and analysis charts.

    Attributes:
        default_parameters: customized strategy attributes.

    """

    def __init__(self, **default_parameters):
        """inits strategy."""
        self.filters = {}
        self._default_parameters = default_parameters
        self.set_parameters(default_parameters)

    def __call__(self, func):
        """decorator function

        Args
          func: A function of customized strategy.
        """
        self.func = func
        return self

    def set_parameters(self, variables):
        """set your customized strategy parameters.

        let strategy class use variables dict to set method.

        Args:
          variables: a dict of your customized strategy attributes.

        """

        # remove stop vars
        stop_vars = ['sl_stop', 'tp_stop', 'ts_stop']
        for svar in stop_vars:
            if hasattr(self, svar):
                delattr(self, svar)

        # set defualt variables
        if self._default_parameters:
            for key, val in self._default_parameters.items():
                setattr(self, key, val)

        # set custom variables
        if variables:
            for key, val in variables.items():
                setattr(self, key, val)

    def show_parameters(self):
        parameters = {}
        for key, val in self._default_parameters.items():
            parameters[key] = getattr(self, key)
        print(parameters)

    @staticmethod
    def _enumerate_filters(ohlcv, filters):
        """enumerate filters data.

        process filter dictionary data to prepare for adding filter signals.

        Args:
          ohlcv: a dataframe of your trading target.
          filters: a dict of your customized filter attributes.

        Returns:
          a dict that generate tuple with filter signal dataframe and figures data.
          for example:

        {'mmi': (timeperiod                    20
          timestamp
          2020-11-25 02:00:00+00:00   true
          2020-11-25 03:00:00+00:00   true
          2020-11-25 04:00:00+00:00   true

          [3 rows x 1 columns], {'figures': {'mmi_index': timestamp
            2020-11-25 02:00:00+00:00    0.7
            2020-11-25 03:00:00+00:00    0.7
            2020-11-25 04:00:00+00:00    0.7
            name: close, length: 28597, dtype: float64}})}

        """
        ret = {}
        for fname, f in filters.items():
            # get filter signals and figures
            filter_df, filter_figures = f(ohlcv)
            ret[fname] = (filter_df, filter_figures)
        return ret

    @staticmethod
    def _add_filters(entries, exits, fig_data, filters):
        """add filters in strategy.

        generate entries, exits, fig_data after add filters.

        Args:
          entries: A dataframe of entries point time series.
          exits: A dataframe of exits point time series.
          fig_data: A dict of your customized figure Attributes.
          filters: A dict of _enumerate_filters function return.

        Returns:
          entries: A dataframe of entries point time series after add filter function.
          exits: A dataframe of exits point time series after add filter function.
          fig_data: A dict of tuple with filter signal dataframe and figures data.

        """
        for fname, (filter_df, filter_figures) in filters.items():
            filter_df.columns = filter_df.columns.set_names([fname + '_' + n for n in filter_df.columns.names])
            entries = filter_df.vbt.tile(entries.shape[1]).vbt & entries.vbt.repeat(filter_df.shape[1]).vbt
            exits = exits.vbt.repeat(filter_df.shape[1])
            exits.columns = entries.columns

            # merge figures
            if filter_figures is not None:
                if 'figures' in filter_figures:
                    if 'figures' not in fig_data:
                        fig_data['figures'] = {}
                    for name, fig in filter_figures['figures'].items():
                        fig_data['figures'][fname + '_' + name] = fig
                if 'overlaps' in filter_figures:
                    if 'overlaps' not in fig_data:
                        fig_data['overlaps'] = {}
                    for name, fig in filter_figures['overlaps'].items():
                        fig_data['overlaps'][fname + '_' + name] = fig

        return entries, exits, fig_data

    @staticmethod
    def _add_stops(ohlcv, entries, exits, variables):
        """Add early trading stop condition in strategy.

        Args:
          ohlcv: A dataframe of your trading target.
          entries: A dataframe of entry point time series.
          exits: A dataframe of exits point time series.
          variables: A dict of your customized strategy Attributes.

        Returns:
          entries: A dataframe of entries point time series after add stop_early function.
          exits: A dataframe of exits point time series after add stop_early function.

        """
        entries, exits = stop_early(ohlcv, entries, exits, variables)
        entries = entries.squeeze()
        exits = exits.squeeze()
        return entries, exits

    def backtest(self, ohlcv, variables=None,
                 filters=None, lookback=None, plot=False,
                 signals=False, side='long', cscv_nbins=10,
                 cscv_objective=lambda r: r.mean(), html=None, compounded=True, execution_price='close', **args):

        """Backtest analysis tool set.
        Use vectorbt as base module to create numerical operations features.
        Use seaborn and pyechart as base modules to create analysis charts platform.

        Args:
          ohlcv: A dataframe of your trading target.
          variables: A dict of your customized strategy Attributes.
            Default is None.
          filters: A dict of your customized filter Attributes.
            Default is None.
          lookback: A int of slice that you want to get recent ohlcv.
            Default is None.
          plot: A bool of control plot display.
            Default is False.
          signals: A bool of controlentries, exits, fig_data return.
            Default is False.
          side: A str of transaction direction,short or long.
            Default is long.
          cscv_nbins: A int of CSCV algorithm bin size to control overfitting calculation.
            Default is 10.
          cscv_objective: A function of in sample(is) and out of sample(oos) return benchmark algorithm.
            Default is lambda r:r.mean().
          html: A str of your customized html format file to show plot.
            Default is None.
          compounded: use compounded return as result of backtesting
            Default is True
          execution_price: price for trading operation ('open' or 'close').
            Default is 'open'
          **args:
            Other parameters.

        Returns:
            A dataframe of vectorbt.Portfolio.from_signals results
            Plot results display.

        Raises:
            'Shorting is not support yet':if side is 'short'.
            "side should be 'long' or 'short'":if side is not 'short' or 'long'.

        """
        variables = variables or dict()
        filters = filters or dict()

        variables_without_stop = copy.copy(variables)
        exit_vars = ['sl_stop', 'ts_stop', 'tp_stop']
        stop_vars = {}
        for e in exit_vars:
            if e in variables_without_stop:
                stop_vars[e] = variables[e]
                variables_without_stop.pop(e)

        ohlcv_lookback = ohlcv.iloc[-lookback:] if lookback else ohlcv

        variable_enumerate = enumerate_variables(variables_without_stop)

        if not variable_enumerate:
            variable_enumerate = [self._default_parameters]

        entries, exits, fig_data = enumerate_signal(ohlcv_lookback, self, variable_enumerate)

        if filters:
            filter_signals = self._enumerate_filters(ohlcv_lookback, filters)
            entries, exits, fig_data = self._add_filters(entries, exits, fig_data, filter_signals)

        entries, exits = self._add_stops(ohlcv_lookback, entries, exits, stop_vars)

        if signals:
            return entries, exits, fig_data

        if side == 'long':

            if not compounded:
                args['size'] = vbt.defaults.portfolio['init_cash'] / ohlcv_lookback.close[0]

            assert execution_price == 'close' or execution_price == 'open'
            # price = ohlcv_lookback[execution_price] if execution_price == 'close' else ohlcv_lookback[execution_price].shift(-1).bfill()

            portfolio = vbt.Portfolio.from_signals(
                ohlcv_lookback[execution_price], entries.fillna(False), exits.fillna(False), **args)

        elif side == 'short':
            raise Exception('Shorting is not support yet')

        else:
            raise Exception("side should be 'long' or 'short'")

        if (plot or html is not None) and isinstance(entries, pd.Series):
            plot_strategy(ohlcv_lookback, entries, exits, portfolio, fig_data, html=html)

        elif plot and isinstance(entries, pd.DataFrame):

            # perform CSCV algorithm
            cscv = CSCV(n_bins=cscv_nbins, objective=cscv_objective)
            cscv.add_daily_returns(portfolio.daily_returns())
            cscv_result = cscv.estimate_overfitting(plot=False)

            # plot results
            plot_combination(portfolio, cscv_result)
            plt.show()
            variable_visualization(portfolio)

        return portfolio



import math
import pandas as pd
import scipy.signal as sg


def sma(price, n):
    return price.rolling(n).mean()


def wma(price, n):
    return price.ewm(com=n).mean()


# Highpass filter by John F. Ehlers, converted by DdlV
def highpass(Data, n=48):
    a = (0.707*2*math.pi) / n

    alpha1 = (math.cos(a)+math.sin(a)-1)/math.cos(a)
    b = 1-alpha1/2
    c = 1-alpha1

    ret = [0] * len(Data)
    for i in range(2, len(Data)):
        ret[i] = b*b*(Data.iloc[i]-2*Data[i-1]+Data.iloc[i-2])+2*c*ret[i-1]-c*c*ret[i-2]
    return pd.Series(ret, index=Data.index)


def lowpass(Data, n):
    a = 2.0/(1+n)
    lp = [Data[0], Data[1]] + [0] * (len(Data) - 2)
    for i in range(2, len(Data)):
        lp[i] = (a - 0.25 * a * a) * Data[i] + 0.5 * a * a * Data[i-1]
        - (a-0.75 * a * a) * Data[i-2]
        + 2 * (1.-a) * lp[i-1]
        - (1.-a) * (1.-a) * lp[i-2]

    return pd.Series(lp, index=Data.index)


def hullma(price, n):
    wma1 = wma(price, n//2)
    wma2 = wma(price, n)
    return wma(wma1 * 2 - wma2, int(math.sqrt(n)))


def zlma(price, n):
    """
    John Ehlers' Zero lag (exponential) moving average
    https://en.wikipedia.org/wiki/Zero_lag_exponential_moving_average
    """
    lag = (n - 1) // 2
    series = 2 * price - price.shift(lag)
    return wma(series, n)


def alma(price, n):
    # triangular window with 60 samples.
    h = sg.gaussian(n, n*0.2)

    # We convolve the signal with this window.
    fil = sg.convolve(price, h / h.sum())
    filtered = pd.Series(fil[:len(price)], index=price.index)
    return filtered


def detrend(price, n):
    return price - highpass(price, n)


def linear_reg(price, n):
    import talib
    return talib.LINEARREG(price, timeperiod=n)


trends = {
    'sma': sma,
    'wma': wma,
    'lowpass': lowpass,
    'hullma': hullma,
    'zlma': zlma,
    'alma': alma,
    'detrend': detrend,
    'linear_reg': linear_reg
}



import os
import numpy as np
import pandas as pd
from finlab_crypto.online import TradingPortfolio


def create_trading_portfolio(trading_methods: list, key=None, secret=None):
    if not (key and secret):
        key = os.environ.get('KEY')
        secret = os.environ.get('SECRET')
    if key and secret:
        tp = TradingPortfolio(key, secret)
    else:
        raise ValueError('No trading KEY or SECRET provided')
    if trading_methods:
        for tm in trading_methods:
            tp.register(tm)
        tp.register_margin('USDT', 1000 * len(trading_methods))
    else:
        print('Please provide trading methods')
    return tp


def concat_assets(ohlcv, symbols, start_bar=2000):
    close_ref = pd.Series(np.concatenate([
        ohlcv[s].close.astype(float).pct_change().values[start_bar:] for s in symbols]))
    ret_close = (close_ref + 1).cumprod()

    # use concatenate and list comprehension
    close = pd.Series(np.concatenate([
        ohlcv[s].close.astype(float).values[start_bar:] for s in symbols]))
    high = pd.Series(np.concatenate([
        ohlcv[s].high.astype(float).values[start_bar:] for s in symbols]))
    low = pd.Series(np.concatenate([
        ohlcv[s].low.astype(float).values[start_bar:] for s in symbols]))
    open_ = pd.Series(np.concatenate([
        ohlcv[s].open.astype(float).values[start_bar:] for s in symbols]))
    volume = pd.Series(np.concatenate([
        ohlcv[s].volume.astype(float).values[start_bar:] for s in symbols]))

    ret_high = ret_close * high / close
    ret_low = ret_close * low / close
    ret_open = ret_close * open_ / close
    # ret_asset = pd.Series(np.concatenate([[s] * (len(ohlcv[s])-start_bar) for s in symbols]))

    # assert len(ret_asset) == len(ret_close)
    min_datetime = min([ohlcv[s][start_bar:].index.min() for s in symbols])
    # index = pd.MultiIndex.from_arrays(
    # [ret_asset, pd.date_range(min_datetime, periods=len(ret_close), freq='4h').tolist()],
    # names=('asset', 'interval'))

    index = pd.date_range(min_datetime, periods=len(ret_close), freq='4h')
    return pd.DataFrame({
        'open': ret_open.values,
        'high': ret_high.values,
        'low': ret_low.values,
        'close': ret_close.values,
        'volume': volume.values,
    }, index=index).dropna()
